# spike normalization for fspma
# ver 0.6
# (C) P. Sykacek 2004, 2005
# new version allows for better normalization of scale.
# we first calculate the scale of all spike genes accross all slides and
# move each existing file to that scale. This is much more sensible to get
# values that are close to the original data than moving the scale to 1. 

# spike based normalization is done by equation based loess
# smoothing. The loess fit is done for every slide (or even grid)
# separately and uses spike genes. These are either true spikes or
# "houskeeping" genes that are not put into the anlysis afterwards.

# important addition: scale is always taken per slide. That is
# even if the normalization is conditional on grid, the scale is
# removed for the entire slide. (Otherwise there are not enough
# samples to estimate that reliably).

# new information components:

#info$xcol
#info$xcolnam
#info$ycol
#info$ycolnam
#info$gridnocol
#info$gridnocolnam
#info$spikegenes
#
# new info entry for normalization method (info$norm.type):
# spike location spatial amplitude grid scale
# explanation:
# spike location - global mean based on spikes
# spike spatial - global spatial model based on spot positions
# spike amplitude - loess M~A fit based on spikes
# spike grid - mean removal based on subgrid number
# remFlg     - an additional modifier that, if present, triggers
#              flaged spikes to be removed from the loess fit.
# furthermore all combinations of the above are possible.
# e.g. spike spatial amplitude grid normalizes based on spatial position and
# amplitude within grids. adding scale removes the corresponding scale.

# The following function that converts info$norm.type into the appropriate
# normalization information.
get.spk.nrm.eqn <- function(info, model.info)
{
  # check info$norm.type and define model.info$spike.eq,
  # model.info$do.grid, model.info$do.scale and
  # model.info$loess.ctrl that we need to normalize
  # the data appropriately.
  if (any(grep('spike', info$norm.type, fixed=T)))
    {
      if (n.spikes <- length(info$spikegenes)==0)
        stop('Error in def file: Spike based normalization requires spike genes.')
      addspike.info <- list(do.grid=F, do.scale=F, rem.Flg=F, spike.eq=M~1,
                            loess.ctrl=-1)
      if (any(loess.dx <- !is.na(as.numeric(info$norm.type))))
        addspike.info$loess.ctrl <- as.list(as.numeric(info$norm.type[loess.dx]))
      if (length(addspike.info$loess.ctrl)>2)
        addspike.info$loess.ctrl <- addspike.info$loess.ctrl[c(1,2)]
      if (length(addspike.info$loess.ctrl)==1)
        addspike.info$loess.ctrl <- c(addspike.info$loess.ctrl, list(1))
      names(addspike.info$loess.ctrl) <- c('span', 'deg')
      if (addspike.info$loess.ctrl$deg<1 | addspike.info$loess.ctrl$deg>2)
        addspike.info$loess.ctrl$deg <- 1
      eq.str <- 'M~'
      n.strs <- length(info$norm.type)
      n <- 1
      while(n<=n.strs)
        {
          if(eq.str=='M~')
            eq.str <- paste(eq.str, switch(info$norm.type[n],
                                             location='1',
                                             spatial='x:y',
                                             amplitude='A',
                                             ''),
                            sep='')
          else
            eq.str <- paste(eq.str, switch(info$norm.type[n],
                                             location='+1',
                                             spatial='+x:y',
                                             amplitude='+A',
                                             ''),
                            sep='')
          if(info$norm.type[n]=='scale')
            addspike.info$do.scale <- T
          if(info$norm.type[n]=='grid')
            addspike.info$do.grid <- T
          if(info$norm.type[n]=='remFlg')
            addspike.info$rem.Flg <- T
          n <- n+1
        }
      #print(eq.str)
      #print(addspike.info)
      if(eq.str=='M~')
        {
          cat('Spike normalization ill defined, assume global location model.\n')
          eq.str <- 'M~1'
        }
      addspike.info$spike.eq <- as.formula(eq.str)
      model.info <- c(model.info, addspike.info)
    }
  model.info
}  

#info <- list(norm.type=c('spike', 'grid', 'spatial', 'amplitude', 'scale', 'location'), spike.genes=c('a', 'b'))
#model.info <- list()
#model.info.n <- get.spk.nrm.eqn(info, model.info)


# new RG components:
# these have the same dimension like channels and flags
#RG$x
#RG$y
#RG$gridno
#RG$slideno - added for easier manipulations of spike based normalization.
#             and in preparation of unbalanced designs

RGgene.2.spikeframe <- function(RG, gridno=NULL, slideno=NULL)
  # convert RG structure (the part of net genes) into a spike norm frame
  {
    # first select slide and grid
    use.dx <- array(rep(T, prod(dim(RG$R))), dim(RG$R))
    if (!is.null(gridno))
      {
        dx.dim <- dim(use.dx)
        use.dx <- use.dx&RG$gridno==gridno
        dim(use.dx) <- dx.dim
      }
    if (!is.null(slideno))
      {
        dx.dim <- dim(use.dx)
        use.dx <- use.dx&RG$slideno==slideno
        dim(use.dx) <- dx.dim
      }
    # then extract net signals
    R.s <- RG$R[use.dx]-RG$Rb[use.dx]
    G.s <- RG$G[use.dx]-RG$Gb[use.dx]
    # and trim values
    R.s[R.s<fspma.eps] <- fspma.eps
    G.s[G.s<fspma.eps] <- fspma.eps
    # get log amplitude and log ratio
    A <- log2(R.s)+log2(G.s)
    M <- log2(R.s)-log2(G.s)
    dim(A) <- NULL
    dim(M) <- NULL
    x <- RG$x[use.dx]
    dim(x) <- NULL
    y <- RG$y[use.dx]
    dim(y) <- NULL
    RG.spike.frame <- data.frame(M=M, A=A, x=x, y=y)
    RG.spike.frame
  }

spikeframe.2.RGgene <- function(RG, spike.frame, gridno=NULL, slideno=NULL)
  # write normalized spike frame data into RG - genes
  {
    # first select slide and grid
    use.dx <- array(rep(T, prod(dim(RG$R))), dim(RG$R))
    if (!is.null(gridno))
      {
        dx.dim <- dim(use.dx)
        use.dx <- use.dx&RG$gridno==gridno
        dim(use.dx) <- dx.dim
      }
    if (!is.null(slideno))
      {
        dx.dim <- dim(use.dx)
        use.dx <- use.dx&RG$slideno==slideno
        dim(use.dx) <- dx.dim
      }
    eA <- 2^(spike.frame$A)
    eM <- 2^(spike.frame$M)
    R <- sqrt(eA*eM)
    G <- sqrt(eA/eM)
    RG$R[use.dx] <- R
    RG$Rb[use.dx] <- 0
    RG$G[use.dx] <- G
    RG$Gb[use.dx] <- 0
    RG
  }

# this is a data frame of spike channels and info.
#RG$normspike
#RG$normspike$R : Cy5 channel foreground
#RG$normspike$Rb : Cy5 channel background
#RG$normspike$G : Cy3 channel foreground
#RG$normspike$Gb : Cy3 channel background
#RG$normspike$Flg: Spot flag, potentially used to exclude spikes from normalization
#RG$normspike$x: x spot position
#RG$normspike$y: y spot position 
#RG$normspike$gridno: spot grid number
#RG$normspike$slideno: slide number (normalization, unbalanced designs)
#RG$normspike$valrt : known ratio of R/G

RGspike.2.spikeframe <- function(RG, info, model.info, gridno=NULL,
                                 slideno=NULL)
  {
    # extracts data frame of spike spots. This function exists in one direction only.
    # The reverse is not necessary, since we need the "normalization" object only for
    # normalizing the RG gene data.
    # first select slide and grid
    # use.dx <- rep(T, length(RG$normspike$R))
    # use only non bad quality marked spots for normalisation
    use.dx <- !is.na(RG$normspike$Flg)
    if (!is.null(gridno))
      use.dx <- use.dx&RG$normspike$gridno==gridno
    if (!is.null(slideno))
      use.dx <- use.dx&RG$normspike$slideno==slideno
    # extract slide and grid number
    slideno <- RG$normspike$slideno[use.dx]
    gridno <- RG$normspike$gridno[use.dx]
    #print(slideno)
    #print(gridno)
    #cat(sum(use.dx), ' genes selected\n')
    #cat(length(use.dx), length(RG$normspike$R), '\n')
    # then extract net signals
    R.s <- RG$normspike$R[use.dx]-RG$normspike$Rb[use.dx]
    G.s <- RG$normspike$G[use.dx]-RG$normspike$Gb[use.dx]
    #print(is.numeric(R.s))
    #print(is.numeric(G.s))
    if (model.info$rem.Flg)
      {
        Flg <- RG$normspike$Flg[use.dx]
        dx.flg.nok <- Flg==info$flag.nack
        R.s[dx.flg.nok] <- NA
      }
    # keep only non NA values
    keep.dx <- !(is.na(R.s) | is.na(G.s))
    R.s <- R.s[keep.dx]
    G.s <- G.s[keep.dx]
    #print(R.s)
    #print(G.s)
    slideno <- slideno[keep.dx]
    dim(slideno) <- NULL
    gridno <- gridno[keep.dx]
    dim(gridno) <- NULL
    # the true ratio expected to be there
    t.rt <- RG$normspike$t.rt[use.dx]
    t.rt <- t.rt[keep.dx]
    # and trim values
    R.s[R.s<fspma.eps] <- fspma.eps
    G.s[G.s<fspma.eps] <- fspma.eps
    # get log amplitude and log ratio
    # cat('came here\n')
    if (length(R.s)>0)
      {
        A <- log2(R.s)+log2(G.s)
        M <- log2(R.s)-log2(G.s)
        # finally subtract the true log ratio:
        # we fit only the residual process.
        # cat('came here\n')
        M <- M - log2(t.rt)
        #cat('came here\n')
        dim(A) <- NULL
        dim(M) <- NULL
        x <- RG$normspike$x[use.dx]
        x <- x[keep.dx]
        dim(x) <- NULL
        y <- RG$normspike$y[use.dx]
        y <- y[keep.dx]
        dim(y) <- NULL
        RG.spike.frame <- data.frame(M=M, A=A, x=x, y=y, slideno=slideno, gridno=gridno, t.rt=t.rt)
      }
    else
      RG.spike.frame <- data.frame(M=c(), A=c(), x=c(), y=c(), slideno=c(), gridno=c())
    #cat('came end\n')
    RG.spike.frame
  }

RGnormalized.spike.2.spikeframe <- function(RG, slideno=NULL, gridno=NULL)
  {
    # extracts data frame of normalized spike spots. This function exists in one direction only.
    # The reverse is not necessary, since we need the "normalization" object only for
    # normalizing the RG gene data.
    # first select slide and grid
    use.dx <- rep(T, length(RG$normalized.spikes$M))
    if (!is.null(slideno))
      use.dx <- use.dx&RG$normalized.spikes$slideno==slideno
    if (!is.null(gridno))
      use.dx <- use.dx&RG$normalized.spikes$gridno==gridno
    M <- RG$normalized.spikes$M[use.dx]
    A <- RG$normalized.spikes$A[use.dx]
    x <- RG$normalized.spikes$x[use.dx]
    y <- RG$normalized.spikes$y[use.dx]
    slideno <- RG$normalized.spikes$slideno[use.dx]
    gridno <- RG$normalized.spikes$gridno[use.dx]
    t.rt <- RG$normalized.spikes$t.rt[use.dx]
    # return that information.
    list(M=M, A=A, x=x, y=y, slideno=slideno, gridno=gridno, t.rt=t.rt)
  }


# next functions: normalize RG based on spike genes
# first we define a function that fits the spike model
# to fit.frame and removes trends from norm.frame

apply.norm <- function(fit.frame, model.eqn, norm.frame, deg=1, span=0.9) #, do.scale)
  {
    if (model.eqn==as.formula(M~1))
      {
        # calculate location
        loc <- mean(fit.frame$M)
        # and remove it
        norm.frame$M <- norm.frame$M-loc
      }
    else
      {
        # get loess object that allows extrapolation
        #cat('fit \n')
        l.obj <- loess(model.eqn, fit.frame, span=span, degree=deg,
                       control = loess.control(surface = "direct"))
        # predict with it
        # cat('predict')
        vals <- predict(l.obj, norm.frame)
        # and remove prediction from both! 
        norm.frame$M <- norm.frame$M-vals
        # (fit.frame might be used for
        # claculating the residual scale!)
        vals <- predict(l.obj, fit.frame)
        fit.frame$M <- fit.frame$M-vals
      }
    # REMARK: scale is handeled outside !
    #if(do.scale)
    #  {
    #    # get global scale
    #    scl <- sd(as.vector(fit.frame$M))
    #    # and divide by it
    #    norm.frame$M <- norm.frame$M/scl
    #  }
    # return normalized frame
    list(norm.frame=norm.frame, fit.frame=fit.frame)
  }

rem.spike.scale <- function(fit.frame, norm.frame, targ.scale)
  # fit.frame is the slide wise spike frame with conditional expectation
  # already removed. (which matters in that order only for loess removal)
  # this is exactly what apply.norm does
  # targ.scale is the target scale we aim at (which is calculated from all
  # spikes of the entire experiment)
  {
    # get (slide) global scale
    scl <- sd(as.vector(fit.frame$M))
    if (scl<fspma.eps)
      scl <- fspma.eps
    # and divide by it
    norm.frame$M <- norm.frame$M*targ.scale/scl
    fit.frame$M <- fit.frame$M*targ.scale/scl
    list(norm.frame=norm.frame, fit.frame=fit.frame)
  }

get.exp.scale <- function(RG, info, model.info)
  # extract sclae for all pikes in the experiment.
  # this is used as target scale to get relations
  # that are close to the original values in the
  # experiment.
  {
    spike.data <- RGspike.2.spikeframe(RG, info, model.info)
    trg.sd <- sd(as.vector(spike.data$M))
    trg.sd
  }

spike.normalize <- function(RG, info, model.info)
  {
    # does slide normalization 
    # model.info$spike.eq - equation for loess fit (or indicator for
    # location based normalization)
    # model.info$do.grid - if T: grid based normalization (remove pin effects)
    # model.info$do.scale - if T: remove scale
    # model.info$$loess.ctrl - loess span parameter
    require(stats)
    # number of slides:
    all.slides <- sort(unique(as.vector(RG$slideno)))
    n.slides <- length(unique(all.slides))
    # number of gridpoints
    if (model.info$do.grid)
      {
        all.grid <- sort(unique(as.vector(RG$gridno)))
        n.grid <- length(all.grid)
        # just a rough estimate of number of spikes per loess fit.
        n.loess.tran <- sum(RG$gridno==all.grid[1] & RG$slideno == all.slides[1])
      }
    else
      {
        all.grid <- NULL
        n.grid <- 0
        # again just a rough estimate
        n.loess.tran <- sum(RG$slideno == all.slides[1])
      }
    # we have loess based normalization
    # first determine whether we need to set loess smoothness constant
    # automatically.
    if (length(model.info$loess.ctrl$span)>1 | model.info$loess.ctrl$span <= 0 |
        model.info$loess.ctrl$span >=1)
      {
        if (n.loess.tran > 1000)
          model.info$loess.ctrl$span <- 0.75
        else
          model.info$loess.ctrl$span <- 0.75 + 0.25*(1-n.loess.tran/1000)
      }
    #print(all.slides)
    # first we get the target scale (to normalize towards that)
    targ.scale <- get.exp.scale(RG, info, model.info)
    normalized.spikes <- c()
    # loop over slides and possibly subgrids
    for(slideno in all.slides)
      {
        cat('.')
        if (model.info$do.grid) # then we do it within subgrids 
          {
            all.fit <- c()
            for(gridno in all.grid)
              {
                cat(':')
                # get data to fit loess
                l.fit.dat <- RGspike.2.spikeframe(RG, info, model.info,
                                                  gridno, slideno)
                if(length(l.fit.dat$M)==0)
                  {
                    print(l.fit.dat)
                    cat('There are no spikes in slide ', slideno, ' subgrid ', gridno, ' execution aborts \n')
                    stop('Processing halted')
                  }
                # get data that will be normalised
                norm.dat <- RGgene.2.spikeframe(RG, gridno, slideno)
                # normalise
                ret <- apply.norm(l.fit.dat, model.info$spike.eq, norm.dat,
                                       model.info$loess.ctrl$deg,  model.info$loess.ctrl$span)
                norm.dat <- ret$norm.frame
                if(length(all.fit)==0)
                  all.fit <- ret$fit.frame
                else
                  all.fit <- rbind(all.fit, ret$fit.frame)
                # and write normalized data back to RG
                RG <- spikeframe.2.RGgene(RG, norm.dat, gridno, slideno)
              }
            # now we may need to normalize for scale
            if (model.info$do.scale)
              {
                # get the entire slide into frame
                norm.dat <- RGgene.2.spikeframe(RG, slideno=slideno)
                # remove within slide scale (based on spikes only)
                ret <- rem.spike.scale(all.fit, norm.dat, targ.scale)
                norm.dat <- ret$norm.frame
                all.fit <- ret$fit.frame
                # and write it back to RG
                RG <- spikeframe.2.RGgene(RG, norm.dat, slideno=slideno)
              }
            if (length(normalized.spikes)==0)
              normalized.spikes <- all.fit
            else
              normalized.spikes <- rbind(normalized.spikes, all.fit)
          }
        else # we loop over slides only - grid is not a regressor
          {
            # get data to fit loess
            l.fit.dat <- RGspike.2.spikeframe(RG, info, model.info,
                                              slideno=slideno)
            if(length(l.fit.dat$M)==0)
              {
                print(l.fit.dat)
                cat('There are no spikes on slide ', slideno, ' execution aborts \n')
                stop('Processing halted')
              }
            # get data that will be normalised
            norm.dat <- RGgene.2.spikeframe(RG, slideno=slideno)
            # normalise for conditional expectation
            ret <- apply.norm(l.fit.dat, model.info$spike.eq, norm.dat,
                                    model.info$loess.ctrl$deg,  model.info$loess.ctrl$span)
            fit.frame <- ret$fit.frame
            if (model.info$do.scale)
              {
                # normalize scale
                ret <- rem.spike.scale(fit.frame, ret$norm.frame, targ.scale)
                norm.dat <- ret$norm.frame
                fit.frame <- ret$fit.frame
              }
            else # take normalized data (scale left unchanged)
              norm.dat <- ret$norm.frame
            # and write normalized data back to RG
            RG <- spikeframe.2.RGgene(RG, norm.dat, slideno=slideno)
            if (length(normalized.spikes)==0)
              normalized.spikes <- ret$fit.frame
            else
              normalized.spikes <- rbind(normalized.spikes, ret$fit.frame)
          }
      }
    cat('\n')
    # attach normalized spikes
    RG <- c(RG, list(normalized.spikes=normalized.spikes))
    # return RG
    RG
  }

# classical normalization without spikes (if data is on log scale)
# these functions replace yasma's corresponding functions for RG
# objects that have been transformed to log2 scale outside fspma.
# (usually these objects are read from tab delimited files like that)

fspma.logRG2MA <- function(RG)
  {
    # mover an RG object with data on log2 scale to MA representation
    # no background here (although that would be forced zero anyway)!
    R <- RG$R
    G <- RG$G
    M <- R-G
    dim(M) <- dim(RG$R)
    A <- R+G
    dim(A) <- dim(RG$R)
    MA <- list(M=M, A=A)
    MA
  }

fspma.MA2logRG <- function(MA, RG)
  {
    # convert MA to R and G values and set those in RG
    M <- MA$M
    A <- MA$A
    R <- (M+A)/2
    dim(R) <- dim(RG$R)
    G <- (A-M)/2
    dim(G) <- dim(RG$G)
    RG$R <- R
    RG$Rb <- array(rep(0, prod(dim(R))), dim(R))
    RG$G <- G
    RG$Gb <- array(rep(0, prod(dim(G))), dim(G))
    RG
  }

fspma.logRG.loess.norm <- function(RG,deg=1, span=0.6, show=T,
                                   show.new=F, flat=F, tolog=F,...)
  # apply loess normalization to RG object that is possibly on a log scale
  # hence we either use yasma's rg2ma or the implementation provided locally
  {
    #cat('loess call: ', deg, span, show, show.new, flat, tolog) 
    if(!tolog)
      MA <- fspma.logRG2MA(RG)
    else
      MA <- rg2ma(RG)
    d <- ncol(MA$A)
    opar <- par(mfrow=c(round(sqrt(d)),ceiling(d/round(sqrt(d)))))
    on.exit(par(opar))
    for (i in 1:d) {
      loe <- loess(MA$M[,i]~MA$A[,i], deg=deg, span=span)
      if (show)
        plot(MA$A[,i],MA$M[,i],pch=".",xlab="1/2 log(R*G)",ylab="log(R/G)",...)
      MA$M[,i] <- MA$M[,i] - predict(loe,MA$A[,i])
      r <- range(MA$A[,i])
      z <- seq(r[1],r[2],length=100)
      if (show) {
        if (!flat) {
          lines(z,predict(loe,z),col="red",lwd=2)
        }
        else  {
          lines(c(0,20),c(0,0),col="red",lwd=2)
        }
        if (show.new)
          points(MA$A[,i],MA$M[,i],pch=".",col="red")
      }
    }
    if(!tolog)
      RG <- fspma.MA2logRG(MA, RG)
    else
      RG <- ma2rg(MA)
    RG
  }

fspma.logRG.shift.norm <- function(RG,method=c("mean","median"), tolog=F) {
  method <- match.arg(method)
  fct <- switch(method,median=median,mean=mean)
  if(!tolog)
    MA <- fspma.logRG2MA(RG)
  else
    MA <- rg2ma(RG)
  s <- apply(MA$M,2,fct)
  MA$M <- t(apply(MA$M,1,function(v) v - s))
  if(!tolog)
    RG <- fspma.MA2logRG(MA, RG)
  else
    RG <- ma2rg(MA)
  RG
}

fspma.logRG.scale.norm <- function(RG,method=c("sd","mad"), tolog=F) {
  method <- match.arg(method)
  fct <- switch(method,sd=sd,mad=mad)
  if(!tolog)
    MA <- fspma.logRG2MA(RG)
  else
    MA <- rg2ma(RG)
  s <- apply(MA$M,2,fct)
  MA$M <- t(apply(MA$M,1,function(v) v/s))
  if(!tolog)
    RG <- fspma.MA2logRG(MA, RG)
  else
    RG <- ma2rg(MA)
  RG
}
