library(yasma)
# This library provides an interface from balanced experimental designs
# to yasma (balanced mixed model ANOVA). The microarray data should be
# organised as sets of tab delimited files (one for every slide) and be
# consitent within experiments (i.e. only one file format per experiment).
# This flexibility is allowed by using a definition file which specifies the
# model and assigns files to each combination of factors that can arise.
# The definition file also allows to specify the way how the data should be
# analsed. It controls imputing values for flaged spots, normalization, and
# the entire analysis (ANOVA table, calculation of variance components and
# listing differentially expressed genes according to comparisons the
# users interest). Model equations are derived automatically from the
# experiments design. One needs for that purpouse only specify the layout of
# the experiment, specify fixed and random effects and the rank effect. If
# contratsta are to be used these have to be specified as well.
# In addition to this interface, the library extends yasma with methods for
# imputing and more general ranking procedures. ANOVA imputation is
# prohibitively complex for most experiments due the memory requirements of
# the underlying linear model and we use k nearest neighbours to impute.
# The ranking procedures alow to analyse general experiments like
# time course data using contrasts and to rank genes by p-values of
# per gene ANOVA models. Since the library is very closely tied in with yasma,
# it is (per definition) under GPL icense.
#
# (C) P. Sykacek 2004, 2005 version 0.4
#
# modifications in version 0.2- 0.4:
# spike based normalisation added, write normalised and imputed data (log ratios
# or log expressions) to tab delimited file (harmonise files),
# allow for unbalanced designs in input - this can apart from normalisation
# at present NOT be processed and is just provided to aid harmonising files.
# Provide column based tab delimited input from one file. (alternative way of
# loading data). Only spike based normalisation is documented for users. The
# remaining part is only for internal use.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA

if (!exists('fspma.eps'))
  fspma.eps <- 10^-10
if (!exists('fspma.knn.max.tol'))
  fspma.knn.max.tol <- 0.001
if(!exists('fspma.logfile'))
  fspma.logfile='fspma.log.txt'
fspma.set.globals <- function(this.eps=10^-10, this.knn.max.tol=0.001,
           this.logfile='fspma.log.txt')
  {
    fspma.eps <- this.eps
    fspma.knn.max.tol <- this.knn.max.tol
    fspma.logfile <- this.logfile
  }
# first we need a function that reads in the expermient design
# Helper function: deternine length of a string
strlen <- function(x)
  {
    if (length(x)>1)
      len <- length(x)
    else
      len <- 1
    l <- 1
    all.len <- NULL
    while (l<=len)
      {
        tmp <- strsplit(x[l], '.')
        this.len <- length(tmp[[1]])
        all.len <- c(all.len, this.len)
        l <- l+1
      }
    all.len
  }

# Helper function: Take a line and extract as many tab delimited
# numbers as are there.
# Pattern: -, decimal point and digits 
# num.pat <- '-?[0-9]+\.[0-9]+|-?[0-9]+|NA'
num.pat <- '-?[0-9]+|NA'
get.numerics <- function(temp.string)
  # get as many numbers as are in temp.string and return a vector of these
  {
    nl <- strlen(temp.string)
    allvals <- NULL
    ret <- 0
    while(nl>0 & ret != -1)
      {
        ret <- regexpr(num.pat, temp.string, perl=T)
        if(ret != -1)
          {
            start <- ret
            dist <- attr(ret, 'match.length')
            # print(start)
            this.strg <- substring(temp.string, start,(start+dist-1))
            allvals <- c(allvals, as.numeric(this.strg))
            temp.string <- substring(temp.string, start+dist, nl)
            nl <- strlen(temp.string)
          }
      }
    allvals
  }

# Next: get all substrings that are delimited by tab and return all but
# the first.
# The first identifies the information, the remaining ones are the values.
get.str.vals <- function(temp.string)
  {
    ret <- strsplit(temp.string, '\t')
    nl <- length(ret)
    ret <- ret[2:nl]
    ret
  }
# we generate the following info:
# info$version: version of def file
# info$name: experiment name
# info$unbalanced: unbalanced experiment -> T or F
# info$effects: number of effects and repetitions within each
# info$effnams: names of effects as used in the yasma equations
# info$randeffs: boolean info about random effects
# info$rank.term: rank term used by anova based ranking
# info$rank.names: rank names used to write results into file
# info$do.anova.ranking: do gene raking using an ANOVA per gene
# info$contrast: contrasts among rank names for testing within ANOVA
# info$ctrfname: output file for contrast based ranking
# info$p.type: type of comparison ('LSD', 'bonf' or 'cnt')
# info$p.thrs: threshold value to cut off significant genes (or number used as 'cnt')
# info$aovfname: output file name for anova result and variance components.
# info$norm.type: control normalization 'NA' if none otherwise 'loess',
#   'location' or 'scale' combinations like c('loess', 'scale') or
#   c('location', 'scale') are allowed as well.
# additional normalisation in version 0.2:
# new info entry for normalization method in version 0.2 (info$norm.type):
# spike location spatial amplitude grid scale the order is irrelevant.
# all is left as strings.
# spike location - global mean based on spikes
# spike spatial - global spatial model based on spot positions
# spike amplitude - loess M~A fit based on spikes
# spike grid - mean removal based on subgrid number
# remFlg     - an additional modifier that, if present, triggers
#              flaged spikes to be removed from the loess fit.
# furthermore all combinations of the above are possible.
# e.g. spike spatial amplitude grid normalizes based on spatial position and
# amplitude within grids. adding scale removes the corresponding scale.
# Function that converts info$norm.type into the appropriate normalization
# information.
# info$dologarray: if T : convert RG using logs, otherwise without log.
# info$skip.hdr: number of rows to skip when reading data files.
# info$flag.nack: value in flags column
# info$impute.mthd: method for imputing flaged spots and NA spots (i.e. missing infoprmation)
# at present: 'del' or 'knn' if the latter an integer that is used as k (i.e. number of closest
# values) must be specified. k is typically between 10 and 20.
# info$load.only: if T : fspma.wrapper exits after loading the data. The flag is
# automatically reset.
# info$genecolnam: name of gene column
# info$rcolnam: column name red
# info$rbcolnam: column name red - background
# info$gcolnam: column name green
# info$gbcolnam: column name green background
# info$flagcolnam: column name flags
# additional in ver 0.2:
# info$Xcolnam: name
# info$Ycolnam: name
# info$gridnocolnam: name
# already version 0.1:
# info$dat.file.rd: list(list(
#                         fnam: file name
#                         expdx: experiment indices -> determine the position in the RG data structure.
#                         doswp: T -> file contains colour swap
#                         ))
# info$geneids: vector of gene names that are read from the file
# info$spikegenes: vector of spike gene names

# all info components
all.inf.nams.v2=c('version', 'name', 'unbalanced', 'effects', 'effnams', 'randeffs', 'rank.term',
  'rank.names', 'do.anova.ranking', 'do.variety', 'contrast', 'ctrfname', 'p.type', 'p.thrs',
  'aovfname', 'norm.type', 'mom.stab', 'dologarray', 'flag.nack', 'impute.mthd',
  'load.only', 'datoutfnam', 'genecolnam', 'rcolnam', 'rbcolnam',
  'gcolnam', 'gbcolnam', 'flagcolnam', 'xcolnam', 'ycolnam', 'gridnocolnam', 'dat.file.rd',
  'geneids', 'spikegenes', 'fast.load', 'gene.infocols', 'gene.infofname')

all.parse.syms.v2=c('Ver:', 'ExpName:', 'Is.Unbalanced:', 'Effects:', 'Eff.Nams:', 'RandEffs:', 'Rank.Eff:',
  'Rank.Names:', 'Base.Contrast:', 'Base.Contrast:', 'Base.Contrast:', 'Contrast.Outfile:', 'Comp.Type:', 'Comp.Thrs:',
  'ANOVA.Outfile:', 'Normalization:', 'Mom.Stab:', 'DoRG2logarray:', 'Flag.NOK:', 'Impute.Mthd:',
  'Load.Only:', 'DatOutFnam:', 'gene.colnam:', 'R.colnam:', 'Rb.colnam:',
  'G.colnam:', 'Gb.colnam:', 'Flag.colnam:', 'X.colnam:', 'Y.colnam:', 'Grid.colnam:', 'file.alloc:',
  'gene.list:', 'spike.list:', 'fast.load:', 'gene.infocols:', 'gene.infofname:')

# those that actually require non NA information
min.inf.req.v2=c('version', 'unbalanced', 'effects', 'effnams', 'randeffs', 'rank.term',
  'rank.names', 'do.anova.ranking', 'do.variety', 'p.type', 'p.thrs', 'aovfname', 'dologarray',
  'load.only', 'genecolnam', 'rcolnam', 'dat.file.rd', 'geneids')

# helper to convert a file combination to an index
# and check that the file combination matches to
# the effects definition
comb2dx <- function(comb, effects)
  {
    nlen <- length(effects)
    effects <- effects[1:(nlen-1)]
    nlen <- nlen-1
    n <- nlen
    dx <- 0
    fkt <- 1
    err <- F
    while (n>0)
      {
        #print(c(effects[n], comb[n]))
        if (effects[n]<comb[n])
          err <- T
        dx <- dx+fkt*(comb[n]-1)
        fkt <- fkt*effects[n]
        n <- n-1
      }
    if (err)
      dx <- NA
    else
      dx <- dx+1
    dx
  }

# inverse function:
dx2comb <- function(dx, effects)
  {
    nlen <- length(effects)
    effects[nlen] <- 1
    n <- 1
    comb <- rep(0, nlen-1)
    fkt <- prod(effects)
    err <- fkt<dx
    if (!err)
      {
        dx <- dx-1
        n <- 1
        while (n<nlen)
          {
            # print(dx)
            comb[n] <- dx %/% prod(effects[(n+1):nlen]) + 1
            dx <- dx %% prod(effects[(n+1):nlen])
            n <- n+1
          }
      }
    else
      comb <- NA
    comb
  }

info.single <- function(info)
  # true for single colour arrays
  {
    (any(info$gcolnam=='NA') & any(info$gbcolnam=='NA'))
  }

have.amp.loess <- function(info)
  # true if there is amplitude based loess in the definition file
  {
    (any(info$norm.type=='loess') & !any(info$norm.type=='spike')) #|
#    (any(info$norm.type=='spike') & any(info$norm.type=='amplitude'))
  }

chk.state.ok = list(req.exist=T, eff.len=T, file.alloc=T, rank.rand=T, rank.coeff=T, design.not.sing=T,
                     norm.affy=T, contrast.len=T, ranknam.len=T, log.vals.bckgrnd=T, impute.ok=T,
                     spike.ok=T, p.admthd.ok=T, p.val.ok=T)

check.info <- function(info)
  {
    # init state with all ok
    chk.state <- chk.state.ok
    all.inf.nams <- all.inf.nams.v2
    min.inf.req <- min.inf.req.v2
    bad.ctr <- c()
    effects.missing <- c()
    problem.files <- c()
    missing.entries <- c()
    # first check whether all components are present (and which are there)
    miss.reqid <- is.na(info[min.inf.req]) | info[min.inf.req]=='NA'
    nshould <- length(min.inf.req)
    missing.entries <- c()
    n <- 1
    while(n <= nshould)
      {
        if (miss.reqid[n])
          {
            id.miss <- grep(min.inf.req[n], all.inf.nams, fixed=T)
            missing.entries <- c(missing.entries, all.parse.syms.v2[id.miss])
          }
        n <- n+1
      }
    chk.state$req.exist <- length(missing.entries)==0
    attr(chk.state$req.exist, 'prs.dx') <- c()
    #print(chk.state$req.exist)
    if (chk.state$req.exist) # we continue only if all necessary entries exist
      {
        # check effect lengths
        chk.state$eff.len <- (length(info$effects) == length(info$randeff)) &&
        (length(info$effects) == length(info$effnams))
        attr(chk.state$eff.len, 'prs.dx') <- c(4,5,6)
        # next: check whether all effect combinations are allocated to files
        neff <- length(info$effects)
        fchk.tab <- rep(F, prod(info$effects[1:(neff-1)]))
        problem.files <- list()
        n.f.alloc <- length(info$dat.file.rd)
        n <- 1
        while(n<=n.f.alloc)
          {
            dx <- comb2dx(info$dat.file.rd[[n]]$expdx, info$effects)
            effects.ok <- !is.na(dx) &&
            is.logical(info$dat.file.rd[[n]]$doswp) &&
            !is.na(info$dat.file.rd[[n]]$doswp) &&
            length(info$dat.file.rd[[n]]$expdx) == length(info $effects)-1
            if (effects.ok)
              fchk.tab[dx] <- T
            else
              problem.files <- c(problem.files, list(info$dat.file.rd[[n]]))
            n <- n+1
          }
        effects.missing <- c()
        maxdx <- prod(info$effects[1:(neff-1)])
        n <- 1
        while(n <= maxdx)
          {
            if (!fchk.tab[n])
              effects.missing <- rbind(effects.missing, dx2comb(n, info$effects))
            n <- n+1
          }
        if (length(effects.missing)>0)
          attr(effects.missing, 'header.names') <- info$effnams[1:(neff-1)]
        chk.state$file.alloc <- (!any(fchk.tab==F) | info$unbalanced) && (length(problem.files)==0)
        attr(chk.state$file.alloc, 'prs.dx') <- c(33)
        attr(chk.state$file.alloc, 'unbalanced') <- info$unbalanced
        if (info$rank.term>0)
          {
            # check whether rank term is not a random effect
            chk.state$rank.rand <- !info$randeff[info$rank.term]
            attr(chk.state$rank.rand, 'prs.dx') <- c(6, 7)
          }
        else
          {
            # check whether rank term zero is only coupled with rank by variety
            chk.state$rank.coeff <- !(any(!is.na(info$contrast)) | info$do.anova.ranking)
            attr(chk.state$rank.coeff, 'prs.dx') <- c(7, 9)
          }
        # check whether the design is not singular
        # this is the case if we have one effect appart replicates on slide that has only one level.
        chk.state$design.not.sing <- sum(info$effects==1)<=1
        attr(chk.state$design.not.sing, 'prs.dx') <- c(4)
        # check whether single channel experiments (affy and the like) have a meaningful
        # normalization (this excludes loess MA smoothers since M==A for Affy and we get a
        # perfect line fit)
        chk.state$norm.affy <- !(info.single(info) & have.amp.loess(info))
        attr(chk.state$norm.affy, 'prs.dx') <- c(16, 24, 26)
        # check whether all contrasts have the correct length.
        if(info$rank.term > 0)
          {
            ncontrasts <- length(info$contrast)
            ctlenreq <- info$effects[info$rank.term]
            n <- 1
            bad.ctr <- list()
            if (!any(is.na(info$contrast)))
              {
                while(n <=ncontrasts)
                  {
                    ct <- info$contrast[[n]]
                    if (length(ct)==1)
                      this.ctok <- (ct > 0) & (ct <= ctlenreq)
                    else
                      this.ctok <- length(ct)==ctlenreq 
                    chk.state$contrast.len <-  chk.state$contrast.len & this.ctok
                    if (!this.ctok)
                      bad.ctr <- c(bad.ctr, info$contrast[n])
                    n <- n+1
                  }
              }
            attr(chk.state$contrast.len, 'prs.dx') <- c(4, 7, 9)
          }
        else # rank term is zero and we need two rank names for the colour difference (if they are used at all)
          ctlenreq <- 2
        # checking whether effect names have the right length
        chk.state$ranknam.len <- ctlenreq == length(info$rank.names)
        attr(chk.state$ranknam.len, 'prs.dx') <- c(4, 7, 8)
        # checking whether we have zero background in case of dologarray false:
        chk.state$log.vals.bckgrnd <- ((info$rbcolnam == 'NA') && (info$gbcolnam == 'NA')) | info$dologarray
        attr(chk.state$log.vals.bckgrnd, 'prs.dx') <- c(18, 25, 27)
        # if there is a flag column and there is no imputing method raise error.
        chk.state$impute.ok <- !((info$flagcolnam != 'NA') & (info$impute.mthd == 'NA'))
        attr(chk.state$impute.ok, 'prs.dx') <- c(28, 20)
        # check whether spikes are there if spike based normalization was selected
        chk.state$spike.ok <- !(any(grep('spike', info$norm.type, fixed=T)) & (length(info$spikegenes)==0 | any(is.na(info$spikegenes))))
        attr(chk.state$spike.ok, 'prs.dx') <- c(16, 34)
        # check whether p-value adjustment is a valid method
        chk.state$p.admthd.ok <- any(info$p.type == p.adjust.methods)
        attr(chk.state$p.admthd.ok, 'prs.dx') <- c(13)
        # check whether p-value is valid (between 0 and 1) or integer
        chk.state$p.val.ok <- ((info$p.thrs > 0) && (info$p.thrs < 1)) || (round(info$p.thrs)==info$p.thrs)
        attr(chk.state$p.val.ok, 'prs.dx') <- c(14)
       }
    list(chk.state=chk.state, bad.ctr=bad.ctr, effects.missing=effects.missing,
         problem.files=problem.files, missing.entries=missing.entries)
  }

stream.log <- function(strg, fname=fspma.logfile, to.wnd=T)
  # write to log stream and default also to stdout.
  {
    fid <- file(fname, "a")
    writeLines(strg, con = fid, sep = "")
    if (to.wnd)
      {
        writeLines(strg, con = stdout(), sep = "")
        # both flush commands... (one for console and thus emacs the other
        # for the Rwin gui.
        flush(con = stdout())
        flush.console() 
      }
    close(fid)
  }

state.err.hdlr <- function(err.name="", prsym.dx=c(), err.lsts)
  {
    if(err.name=="req.exist") # info message for missing entries
      {
        m.e <- err.lsts$missing.entries
        err.strg <- sprintf("Following file entries must be specified, they are however missing or NA:\n %s\n",
                            paste("", m.e, sep="\n", collapse=""))
      }
    else if(err.name=="eff.len") # info message for effect length missmatch
      {
        err.strg <- sprintf("All experimental effect definitions must have identical length.\n")
      }
    else if(err.name=="file.alloc") # info message for file to effect allocation problems
      {
        err.strg <- ""
        unbalanced <- attr(err.lsts$chk.state$file.alloc, "unbalanced")
        e.m <- err.lsts$effects.missing
        if(length(e.m)>0 & ! unbalanced) # missing allocations not flaged as unbalanced design
          {
            err.strg <- sprintf("The effects description is unbalanced without setting the corresponding flag.\n")
            err.strg <- paste(err.strg, "The following effects have no files allocated to:\n", sep="", collapse="")
            #n.cls <- ncol(e.m)
            e.nams <- attr(e.m, 'header.names')
            err.strg <- paste(err.strg, paste("", e.nams, sep="\t", collapse=""), sep="\n", collapse="")
            err.strg <- paste(err.strg, "\n", sep="", collapse="")
            for (ne in 1:nrow(e.m))
              {
              
                err.strg <- paste(err.strg, paste("", e.m[ne,], sep="\t", collapse=""), sep="\n", collapse="")
              }
            err.strg <- paste(err.strg, "\n", sep="", collapse="")
          }
        p.f <- err.lsts$problem.files
        if (length(p.f)>0) # files with wrong effect or dye swap flag entries
          {
            err.strg <- paste(err.strg,"File allocations with wrong effect or dye swap specifier:\n",
                              sep="", collapse="")
            for (f.n in 1:length(p.f))
              {
                f.name <- p.f[[f.n]]$fnam
                eff.d <- p.f[[f.n]]$expdx
                swp <- p.f[[f.n]]$doswp
                add.str <- paste(paste(f.name, paste("", eff.d, sep="\t", collapse="")),
                                 swp, sep="\t", collapse="")
                err.strg <- paste(err.strg, add.str, sep="\n", collapse="")
              }
            err.strg <- paste(err.strg, "\n", sep="", collapse="")
          }
      }
    else if(err.name=="rank.rand") # info message for rank effect indicated to be a random effect
      {
        err.strg <- "Gene ranking can not be done over a random effect\n"
      }
    else if(err.name=="rank.coeff") # info message for rank effect zero and ranking is appemted by contrast or ANOVA
      {
        err.strg <- "Rank effect zero can only be combined with VARIETY (significant average channel difference).\n"
      }
    else if(err.name=="design.not.sing") # info message for singular design
      {
        err.strg <- "This experimental description has a singular solution.\n Only the within slide replicates may have one level.\n"
      }
    else if(err.name=="norm.affy") # info message for normalization problem 
      {
        err.strg <- "Single channel input must not be normalized by an M/A loess fit. (all expressions will be zero)\n"
      }
    else if(err.name=="contrast.len") # info message for wrong contrst length
      {
        err.strg <- "Length of contrast definition does not match number of levels in rank effect\nor baselevel out of range. Possibly missing contrast name?\n"
        bad.ctr <-  err.lsts$bad.ctr
        nlen <- length(bad.ctr)
        for (n in 1:nlen)
          {
            ct.nam <- names(bad.ctr[n])
            add.strg <- paste(ct.nam, paste("", bad.ctr[[n]], sep="\t", collapse=""), sep="", collapse="")
            err.strg <- paste(err.strg, add.strg, sep="", collapse="")
            err.strg <- paste(err.strg, "\n", sep="", collapse="")
          }
      }
    else if(err.name=="ranknam.len") # info message for wrong number of rank names
      {
        err.strg <- "Number of rank names does not match number of levels in rank effect.\n"
      }
    else if(err.name=="log.vals.bckgrnd") # info message for background signals despite missing log transformation
      {
        err.strg <- "Background signals can not be used without conversion to log2 domain.\n"
      }
    else if(err.name=="impute.ok") # info message for missing imputation method despite flag channel specification
      {
        err.strg <- "Bad quality flags can only be used in connection with imputing.\n"
      }
    else if(err.name=="spike.ok") # info message for missing list f spike genes despite spike based normalization
      {
        err.strg <- "Spike based normalisation requires specification of spike genes.\n"
      }
    else if(err.name=="p.admthd.ok") # info message for wrong p-value transformation
      {
        err.strg <- sprintf("Valid p-value adjustments are: %s.\n", paste("", p.adjust.methods, sep=" ", collapse=""))
      }
    else if(err.name=="p.val.ok") # info message for wrong p-value definition.
      {
        err.strg <- "Valid p-value thresholds must be between 0 and 1 or integer, the latter are interpreted as counts.\n"
      }
    if (any(prsym.dx)) # info message which entries in the definition file this error message concerns.
      {
        if (length(prsym.dx)==1)
          fld.strg=sprintf("This problem is caused by inconsistencies in definition file entry: %s.\n\n",
            all.parse.syms.v2[prsym.dx])
        else
          fld.strg=sprintf("This problem is caused by inconsistencies in definition file entries: %s%s.\n\n",
            all.parse.syms.v2[prsym.dx[1]],
            paste("", all.parse.syms.v2[prsym.dx[-1]], sep=", ", collapse=""))
        err.strg <- paste(err.strg, fld.strg, sep="", collapse="\n")
      }
    err.strg
  }

debug.deffile <- function(inf.chk, deffile="", dbg.lognam=fspma.logfile)
  {
    chk.state <- inf.chk$chk.state
    chk.names <- names(chk.state)
    # print(chk.state)
    # print(chk.names)
    if (any(!unlist(chk.state)))
      {
        stream.log(sprintf("Checking the definition file %s indicates the following problems:\n", deffile), dbg.lognam)
        # go through chk.state and generate log file debug entries
        nstates <- length(chk.state)
        n <- 1
        err.cntr <- 1
        while (n <= nstates)
          {
            if (!chk.state[[n]]) # then we have an error in that state and generate an error message
              {
                stream.log(sprintf("Problem number %3.0f:\n", err.cntr), dbg.lognam)
                err.cntr <- err.cntr+1
                #print(chk.names[n])
                err.str <- state.err.hdlr(chk.names[n], attr(chk.state[[n]], 'prs.dx'), inf.chk)
                stream.log(err.str, dbg.lognam)
                #stream.log("\n\n", dbg.lognam)
              }
            n <- n+1
          }
        debug.ok <- F
      }
    else
      {
        stream.log(sprintf("Checking consistency of the definition file %s was successful.\n", deffile),
                   dbg.lognam)
        debug.ok <- T
      }
    debug.ok
  }

fspma.check.deffile <- function(info, deffile="", dbg.lognam=fspma.logfile)
  {
    # interface to check definition files for consistency.
    chk.stat <- check.info(info)
    debug.deffile(chk.stat, deffile, dbg.lognam)
  }

# helper checks info after generation 
#chk.def <- function(info)
#  {
#    print('checking consitency of definition file')
#    # determine version of input file
#    #if (info$version==1)
#    #  {
#    #    all.inf.nams <- all.inf.nams.v1
#    #    min.inf.req <- min.inf.req.v1
#    #  }
#    #else # at present: only 2 versions.
#    #  {
#        all.inf.nams <- all.inf.nams.v2
#        min.inf.req <- min.inf.req.v2        
#    #  }
#    # first check whether all components are present
#    exist.nams <- names(info)
#    nshould <- length(all.inf.nams)
#    nis <- length(exist.nams)
#    all.exist <- (nis==nshould)
#    n <- 1
#    while(n <= nshould & all.exist)
#      {
#        all.exist <- all.exist && any(grep(all.inf.nams[n], exist.nams, fixed=T))
#        n <- n+1
#      }
#    if (!all.exist)
#      {
#        print(all.inf.nams)
#        print(exist.nams)
#        cat('Definition file error: not all info components of version ',
#                 info$version, ' present')
#        stop('Aborting execution')
#      }
#    # check effects
#    effects.ok <- length(info$effects) == length(info$randeff)
#    # now check whether all effects are assigned to files
#    if (!effects.ok)
#      stop('Effects and Random Effects must have equivalent length')
#    # next: check whether all effect combinations are allocated to files
#    neff <- length(info$effects)
#    fchk.tab <- rep(F, prod(info$effects[1:(neff-1)]))
#    n.f.alloc <- length(info$dat.file.rd)
#    n <- 1
#    while(n<=n.f.alloc && effects.ok)
#      {
#        dx <- comb2dx(info$dat.file.rd[[n]]$expdx, info$effects)
#        effekts.ok <- !is.na(dx) & is.logical(info$dat.file.rd[[n]]$doswp)
#        if (effects.ok)
#          fchk.tab[dx] <- T
#        n <- n+1
#      }
#    effects.ok <- effects.ok && !any(fchk.tab==F)
#    #print(info$unbalanced)
#    #print(effects.ok)
#    #print(sum(fchk.tab))
#    #print(length(fchk.tab))
#    if (!effects.ok & !info$unbalanced)
#      stop('Problem with file allocation! This is an unbalanced design')
#    # finaly check whether all required information is there
#    if (sum(these <- is.na(info[min.inf.req]) | info[min.inf.req]=='NA')>0)
#      stop( paste('Missing:', paste('info$', min.inf.req[these], sep='', collapse=' ')))
#    if (info$rank.term>0)
#      {
#        if (info$randeff[info$rank.term])
#          stop('Gene ranking is only possible over fixed effects. Rand.Eff: at pos Rank.Eff: must be F!')
#      }
#    else
#      {
#        if (any(!is.na(info$contrast)) | info$do.anova.ranking)
#          stop('rank.term < 1 can only rank for variety')
#      }
#    if(sum(info$effects==1)>1)
#      stop('Only within slide replicates may have 1 level. This design has only singular solutions.')
#    if (info.single(info) & have.amp.loess(info))
#      stop('Major inconsistency in definition file: M/A loess normalization does not work for single channel specifications. Analysis aborts.')
#        # if there is a flag column and there is no imputing method raise error.
#    if((info$flagcolnam != 'NA') & (info$impute.mthd == 'NA'))
#      stop('Specifying a flag column requires to specify an imputation method. Processing stoped.') 
#    # check whether spikes are there if spike based normalization was selected
#    if (any(grep(info$norm.type, 'spike', fixed=T)) & length(info$spikegenes)==0)
#      stop('Spike based normalisation requires to specify spike genes. Execution aborts.')
#    print('Definition file consitency OK!')
#  }

effect.sort <- function(info)
  # sort effects such that innermost effects are fixed (to guarantee that the p-value calculation
  # is correct.)
  {
    n.effs <- length(info$randeffs)
    effdx <- c(1:n.effs)
    randdx <- effdx[info$randeffs]
    dim(randdx) <- c(length(randdx),1)
    fixdx <- effdx[!info$randeffs]
    dim(fixdx) <- c(1, length(fixdx)) 
    if (any(randdx[,rep(1, length(fixdx))] < fixdx[rep(1, length(randdx)),])) # we need to rearrange
      {
        # first transpose fixdx
        fixdx <- t(fixdx)
        newdx <- rbind(fixdx, randdx)
        # now we rearrange the effected components
        info$randeffs <- info$randeffs[newdx]
        info$effects <- info$effects[newdx]
        info$effnams <- info$effnams[newdx]
        # the new rank term is the position of the old rank term in the new index
        # effdx is just the counter 1:length(effects)
        info$rank.term <- effdx[newdx==info$rank.term]
        # finally we need to do the rearrangement for all allocation entries in
        # info$dat.file.rd
        filedx <- newdx[1:(length(newdx)-1)] # one less since the effect on slide
                                             # replication does not appear here
        n.files <- length(info$dat.file.rd)
        n <- 1
        while(n <= n.files)
          {
            info$dat.file.rd[[n]]$expdx <- info$dat.file.rd[[n]]$expdx[filedx]
            n <- n+1
          }
      }
    info
  }

# helper tests any pattern:

is.patt <- function(pattern, x)
  {
    bool.ret <- regexpr(pattern, x, perl=T)!=-1
    bool.ret
  }

# generate blank string of any length
gen.blnks <- function(nblnks)
  {
    paste('', rep(' ', nblnks), sep='', collapse='')
  }
# helper to split line into substrings with some extra functionality
split.line <- function(pattern, x)
  # split x into tab delimited substrings and
  # remove pattern from list. Removes any empty
  # strings as well.
  {
    strs <- strsplit(x, '\t', fixed=T)
    nlen <- length(strs[[1]])
    n <- 1
    ostrs <- NULL
    while(n<=nlen)
      {
        this.string <- strs[[1]][n]
        # print(this.string)
        res1 <- regexpr(pattern, this.string)
        this.len <- strlen(this.string)
        # print(this.len)
        blankstr <- gen.blnks(this.len)
        # print(res1)
        if (res1 == -1 && this.len!=0 && this.string!=blankstr)
           # not pattern and not emptynot blank
          ostrs <- c(ostrs, this.string)
        n <- n+1
      }
    ostrs
  }

# we generate the following info:
# info$version: version of def file
# info$name: experiment name
# info$unbalanced: unbalanced experiment -> T or F
# info$effects: number of effects and repetitions within each
# info$effnams: names of effects as used in the yasma equations
# info$randeffs: boolean info about random effects
# info$rank.term: rank term used by anova based ranking
# info$rank.names: rank names used to write results into file
# info$do.anova.ranking: do gene raking using an ANOVA per gene
# info$contrast: contrasts among rank names for testing within ANOVA
# info$ctrfname: output file for contrast based ranking
# info$p.type: type of comparison ('LSD', 'bonf' or 'cnt')
# info$p.thrs: threshold value to cut off significant genes (or number if p.type='cnt'
# info$aovfname: output file name for anova result and variance components.
# info$norm.type: control normalization 'NA' if none otherwise 'loess',
#   'location' or 'scale' combinations like c('loess', 'scale') or
#   c('location', 'scale') are allowed as well.
# additional normalisation in version 0.2:
# new info entry for normalization method in version 0.2 (info$norm.type):
# spike location spatial amplitude grid scale the order is irrelevant.
# all is left as strings.
# spike location - global mean based on spikes
# spike spatial - global spatial model based on spot positions
# spike amplitude - loess M~A fit based on spikes
# spike grid - mean removal based on subgrid number
# remFlg     - an additional modifier that, if present, triggers
#              flaged spikes to be removed from the loess fit.
# furthermore all combinations of the above are possible.
# e.g. spike spatial amplitude grid normalizes based on spatial position and
# amplitude within grids. adding scale removes the corresponding scale.
# Function that converts info$norm.type into the appropriate normalization
# information.
# info$dologarray: if T : convert RG using logs, otherwise without log.
# info$skip.hdr: number of rows to skip when reading data files.
# info$flag.nack: value in flags column
# info$impute.mthd: method for imputing flaged spots and NA spots (i.e. missing infoprmation)
# at present: 'del' or 'knn' if the latter an integer that is used as k (i.e. number of closest
# values) must be specified. k is typically between 10 and 20.
# info$load.only: if T : fspma.wrapper exits after loading the data. The flag is
# automatically reset.
# info$genecolnam: name of gene column
# info$rcolnam: column name red
# info$rbcolnam: column name red - background
# info$gcolnam: column name green
# info$gbcolnam: column name green background
# info$flagcolnam: column name flags
# additional in ver 0.2:
# info$Xcolnam: name
# info$Ycolnam: name
# info$gridnocolnam: name
# already version 0.1:
# info$dat.file.rd: list(list(
#                         fnam: file name
#                         expdx: experiment indices -> determine the position in the RG data structure.
#                         doswp: T -> file contains colour swap
#                         ))
# info$geneids: vector of gene names that are read from the file
# info$spikegenes: vector of spike gene names

# for that we specify several regular expressions that are required
# to parse the definition file.
rgx.comment <- '#.*'                    # allows to include comment lines in the def file
rgx.version <- 'Ver:.*'                 # Version number
rgx.expnam <- 'ExpName:.*'              # experiment name
rgx.unbalanced <- 'Is.Unbalanced:.*'    # T : experiment is unbalanced.
rgx.effects <- 'Effects:.*'             # effects row (i.e. number of effects)
rgx.effnams <- 'Eff.Nams:.*'            # effect names (used in ANOVA eqns etc.)
rgx.randeffs <- 'RandEffs:.*'           # rand effects row (which are random)
rgx.rank.term <- 'Rank.Eff:.*'          # rank effect (over which should ranking be done)
rgx.rank.nams <- 'Rank.Names:.*'        # provide rank names.
rgx.contrast <- 'Base.Contrast:.*'      # specify base level or contrast among rank names for within ANOVA test.
rgx.ctrfname <- 'Contrast.Outfile:.*'   # results file name for contrast based comparison
rgx.comp.type <- 'Comp.Type:.*'         # type of p-value evaluation (bonf, lsd, cnt)
rgx.comp.pthrs <- 'Comp.Thrs:.*'        # p-value threshold
rgx.aovfname <- 'ANOVA.Outfile:.*'      # ANOVA and REML output file name
rgx.norm.type <- 'Normalization:.*'     # control slide normalization ('NA', 'loess', 'location', 'scale')
rgx.mom.stab <- 'Mom.Stab:.*'           # control moment stabilization (removes mean and std. dev per gene
                                        # and adds it again before writing the rank tables).
rgx.dologarray <- 'DoRG2logarray:.*'    # control use of RG to array conversion (T or F).
rgx.flagctrl <- 'Flag.NOK:.*'           # flag not ok value (NA if none)
rgx.impute.mthd <- 'Impute.Mthd:.*'     # method for imputing values (del for delete genes,
                                        # knn<TAB>k for knn with k nearest neighbours or NA if none)
rgx.load.only <- 'Load.Only:.*'         # if T : only load data and exit.
rgx.datoutfnam <- 'DataOutFnam:.*'      # name of output file base (2, may be 3 files!)
#rgx.hdr.ctrl <- 'Skip.Hdr:.*'           # number of header lines to skip
#rgx.gencol <- 'gene.col:.*'             # gene column number
rgx.gcnam <- 'gene.colnam:.*'           # gene column name
#rgx.rcol <- 'R.col:.*'                  # Cy5 column number
rgx.rcolnam <- 'R.colnam:.*'            # Cy5 column name
#rgx.rbcol <- 'Rb.col:.*'                # Cy5 background column number (NA if none)
rgx.rbcolnam <- 'Rb.colnam:.*'          # Cy5 background column name (NA if none) 
#rgx.gcol <- 'G.col:.*'                  # Cy3 column number (NA if none -> Affy!)
rgx.gcolnam <- 'G.colnam:.*'            # Cy3 column name (NA of none)
#rgx.gbcol <- 'Gb.col:.*'                # Cy3 background number (NA if none)
rgx.gbcolnam <- 'Gb.colnam:.*'          # Cy3 background column name (NA if none)
#rgx.fcol <- 'Flag.col:.*'               # Flag column number (NA if none)
rgx.fcolnam <- 'Flag.colnam:.*'         # Flag column name (NA if none)
#rgx.xcol <- 'X.col:.*'                  # X-pos column number (NA if none)
rgx.xcolnam <- 'X.colnam:.*'            # X-pos column name (NA if none)
#rgx.ycol <- 'Y.col:.*'                  # Y-pos column number (NA if none)
rgx.ycolnam <- 'Y.colnam:.*'            # Y-pos column name (NA if none)
#rgx.gridnocol <- 'Grid.col:.*'          # Grid no. column number (NA if none)
rgx.gridnocolnam <- 'Grid.colnam:.*'    # Grid.no. column name (NA if none)
rgx.geneinfocols <- 'gene.infocols:.*'  # column names for gene information (NA if none)
rgx.geneinfoname <- 'gene.infofname:.*' # gene information file name (NA if none)
rgx.fileac <- 'file.alloc:.*'           # file allocation list starter (number
                                        # determined from Effects) 
rgx.genelist <- 'genes.list:.*'         # gene list starter (number until something else or eof)
rgx.spikelist <- 'spike.list:.*'        # spike list starter (number until something else or eof)
rgx.fastload <- 'fast.load:.*'          # fast loading of data files. (assumes identical gene positions in all files)
rgx.loadfast <- 'load.fast:.*'          # fast loading of data files. (assumes identical gene positions in all files)

read.defs <- function(fname, log.fname=fspma.logfile, init.logfile=T, print.logheader=T)
  # reads in the definition file that we use to generate the ANOVA evaluation.
  # generates model.info, geneids, experiments and a file list.
  # The actual reading of the data and constriction of the RG structure is
  # done elsewhere.
  # 3 input params.
  # fname:          is the name of the definition file.
  # log.fname:      optional name of the log file (used to store debug information) defaults to fspma.logfile
  # init.logfile:   optional flag, defaults TRUE, in which case we reset the log file (otherwise we append)
  {
    if (init.logfile)
      {
        fid <- file(log.fname, "w")
        close(fid)
      }
    if (print.logheader)
      {
        strg <- paste(paste("Run started on - ", format(Sys.time(), "%a %b %d %Y %X"), sep="", collapse=""),
                      "\n\n", sep="", collapse="")
        stream.log(strg, fname=log.fname)
      }
    # first read definition file
    tmp.def <- scan(fname, what='', sep='\n')
    n.lines <- length(tmp.def)
    # init info:
    info <- as.list(rep(NA, length(all.inf.nams.v2)))
    names(info) <- all.inf.nams.v2
    # print(info$spikegenes)
    # info$flagcol <- NA
    # info$flagcolnam <- NA
    # info$dat.file.rd <- NA
    #  list(list(
    #                         fnam: file name
    #                         expdx: experiment indices -> determine the position 
    #                         ))
    # info$geneids <- NA
    # init state machine allocating files to effects and reading
    # genes is a multi line operation that requires us to use states 
    f2eff.alloc <- F
    gene.lst <- F
    info$do.anova.ranking <- F
    info$do.variety <- F
    n <- 1
    # n.lines <- 100
    while (n <=n.lines)
      {
        # get current line
        curr.str <- tmp.def[n]
        # print(curr.str)
        if (is.patt(rgx.comment, curr.str))
          {
          # ignore a comment line
          }
        else
          {
            if (is.patt(rgx.fastload, curr.str)|is.patt(rgx.loadfast, curr.str))
                                        # get fast load indicator.
              {
                info$fast.load <- as.logical(split.line(rgx.fastload, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }

            if (is.patt(rgx.version, curr.str))
                                        # get experiment version
              {
                info$version <- as.numeric(split.line(rgx.version, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if (is.patt(rgx.expnam, curr.str))
                                        # get experiment name
              {
                info$name <- split.line(rgx.expnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if (is.patt(rgx.unbalanced, curr.str))
                                        # get unbalanced flag
              {
                info$unbalanced <- as.logical(split.line(rgx.unbalanced, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if (is.patt(rgx.effects, curr.str))
                                        # get effects
              {
                info$effects <- get.numerics(curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if (is.patt(rgx.effnams, curr.str))
                                        # get effects
              {
                info$effnams <- split.line(rgx.effnams, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.randeffs, curr.str))
                                        # get random effect info
              {
                info$randeffs <- as.logical(split.line(rgx.randeffs, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.rank.term, curr.str))
                                        # get rank term
              {
                info$rank.term <- get.numerics(curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.rank.nams, curr.str))
                                        # get rank names
              {
                info$rank.names <- split.line(rgx.rank.nams, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.contrast, curr.str))
                                        # get within rank name contrasts for test
              {
                rank.string <- split.line(rgx.contrast, curr.str)
                    # check whether ANOVA ranking is intended
                do.anova.ranking <- any(rank.string=='ANOVA')
                do.variety <- any(rank.string=='VARIETY')
                if (!(do.anova.ranking | do.variety))
                  {
                    # get name of this contrast
                    contrast.name <- rank.string[1]
                    # get numerical values of contrast
                    n.strs <- length(rank.string)
                    contrast <- as.numeric(rank.string[2:n.strs])
                    # process for correct sum if length > 1
                    if (length(contrast) > 1 && sum(contrast)!=0)
                      {
                        sum.p <- sum(contrast[contrast>0])
                        contrast[contrast>0] <- contrast[contrast>0]/sum.p
                        sum.n <- abs(sum(contrast[contrast<0]))
                        contrast[contrast<0] <- contrast[contrast<0]/sum.n
                      }
                    # finally attach it to info$contrast
                    if (any(is.na(info$contrast))) # then it is the first one
                      {
                        info$contrast <- list(contrast)
                        names(info$contrast) <- contrast.name
                      }
                    else
                      {
                        n.contr <- list(contrast)
                        names(n.contr) <- contrast.name
                        info$contrast <- c(info$contrast, n.contr)
                      }
                  }
                if(do.anova.ranking)
                    info$do.anova.ranking <- T
                if(do.variety)
                  info$do.variety <- T
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
           if(is.patt(rgx.ctrfname, curr.str))
                                        # get output filename of contrast based comparison
              {
                info$ctrfname <- split.line(rgx.ctrfname, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.comp.type, curr.str))
                                        # get type of threshold comparison
              {
                info$p.type <- split.line(rgx.comp.type, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.comp.pthrs, curr.str))
                                        # get p-vaue threshold
              {
                info$p.thrs <- as.numeric(split.line(rgx.comp.pthrs, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.aovfname, curr.str))
                                        # get anova and reml results output filename
              {
                info$aovfname <- split.line(rgx.aovfname, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.norm.type, curr.str))
                                        # get normalization control
              {
                info$norm.type <- split.line(rgx.norm.type, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.mom.stab, curr.str))
                                        # moment removal for numerical stability?
              {
                info$mom.stab <- as.logical(split.line(rgx.mom.stab, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.dologarray, curr.str))
                                        # get RG2array control
              {
                info$dologarray <- as.logical(split.line(rgx.dologarray, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.flagctrl, curr.str))
                                        # get NACK flag value
              {
                info$flag.nack <- split.line(rgx.flagctrl, curr.str)
                flag.numerics <- as.numeric(info$flag.nack)
                if (any(info$flag.nack=='NA') | !any(is.na(flag.numerics)))
                  # then we have a numerical flag value
                  info$flag.nack <- flag.numerics
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.impute.mthd, curr.str))
                                        # get method to impute missing values
              {
                info$impute.mthd <- as.list(split.line(rgx.impute.mthd, curr.str))
                # print(info$impute.mthd)
                if (info$impute.mthd[[1]]=='knn')
                  info$impute.mthd[[2]]=as.numeric(info$impute.mthd[[2]])
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.load.only, curr.str))
                                        # get flag whether function exits after data loading
              {
                info$load.only <- as.logical(split.line(rgx.load.only, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.datoutfnam, curr.str))
                                        # get flag whether function exits after data loading
              {
                info$datoutfnam <- split.line(rgx.datoutfnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.hdr.ctrl, curr.str))
#                                        # get header length in data files
#              {
#                info$skip.hdr <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
#            if(is.patt(rgx.gencol, curr.str))
#                                        # get gene columnnumber
#              {
#                info$genecol<- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.gcnam, curr.str))
                                        # get gene column name
              {
                info$genecolnam <- split.line(rgx.gcnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.rcol, curr.str))
#                                        # get Cy5 column number
#              {
#                info$rcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.rcolnam, curr.str))
                                        # get Cy5 column name
              {
                info$rcolnam <- split.line(rgx.rcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.rbcol, curr.str))
#                                        # get Cy5 backgroung column number
#              {
#                info$rbcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.rbcolnam, curr.str))
                                        # get Cy5 background column name
              {
                info$rbcolnam <- split.line(rgx.rbcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.gcol, curr.str))
#                                        # get Cy3 column number
#              {
#                info$gcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.gcolnam, curr.str))
                                        # get Cy3 olumn name
              {
                info$gcolnam <- split.line(rgx.gcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.gbcol, curr.str))
#                                        # get Cy3 background column number
#              {
#                info$gbcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.gbcolnam, curr.str))
                                        # get Cy3 background column name
              {
                info$gbcolnam <- split.line(rgx.gbcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.fcol, curr.str))
#                                        # get flags column number
#              {
#                info$flagcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.fcolnam, curr.str))
                                        # get flags column name
              {
                info$flagcolnam <- split.line(rgx.fcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#ver0.2
#            if(is.patt(rgx.xcol, curr.str))
#                                        # x column
#              {
#                info$xcol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.xcolnam, curr.str))
                                        # x column name
              {
                info$xcolnam <- split.line(rgx.xcolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.ycol, curr.str))
#                                        # y column
#              {
#                info$ycol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.ycolnam, curr.str))
                                        # y column name
              {
                info$ycolnam <- split.line(rgx.ycolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
#            if(is.patt(rgx.gridnocol, curr.str))
#                                        # grid column
#              {
#                info$gridnocol <- get.numerics(curr.str)
#                f2eff.alloc <- F
#                gene.lst <- F
#                spike.lst <- F
#              }
            if(is.patt(rgx.gridnocolnam, curr.str))
                                        # grid column name
              {
                info$gridnocolnam <- split.line(rgx.gridnocolnam, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.geneinfocols, curr.str))
                                        # gene info column names
              {
                info$geneinfocols <-  as.list(split.line(rgx.geneinfocols, curr.str))
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
            if(is.patt(rgx.geneinfoname, curr.str))
                                        # gene info column names
              {
                info$geneinfoname <-  split.line(rgx.geneinfoname, curr.str)
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- F
              }
# end ver0.2
            if(is.patt(rgx.fileac, curr.str))
                                        # file allocation list starting
              {
                f2eff.alloc <- T
                gene.lst <- F
                spike.lst <- F
                first.set <- T
              }
            if(is.patt(rgx.genelist, curr.str))
                                        # gene list starting
              {
                f2eff.alloc <- F
                gene.lst <- T
                spike.lst <- F
                first.set <- T
              }
            if(is.patt(rgx.spikelist, curr.str))
              {                         # spike list starting
                f2eff.alloc <- F
                gene.lst <- F
                spike.lst <- T
                first.set <- T
              }
            if(f2eff.alloc)
              {
                if(first.set)
                  {
                    first.set <- F
                    info$dat.file.rd <- list() # empty list
                  }
                else
                  {
                                        # process line and add it to file allocation
                    tmp.splt <- split.line(rgx.fileac, curr.str)
                    nsplt <- length(tmp.splt)
                    add2lst <- list(tmp.splt[1], as.numeric(tmp.splt[2:(nsplt-1)]),
                                    as.logical(tmp.splt[nsplt]))
                    names(add2lst) <- c('fnam', 'expdx', 'doswp')
                    info$dat.file.rd <- c(info$dat.file.rd, list(add2lst))
                  }
              }
            if(gene.lst)
              {
                if(first.set)
                  {
                    first.set <- F
                    info$geneids <- c() # empty vector
                  }
                else
                                        # add line (trimmed for trailing white space) to gene list
                  {
                    info$geneids <- c(info$geneids, sub('[[:space:]]+$', '', curr.str))
                  }
              }
            if(spike.lst)
              {
                if(first.set)
                  {
                    first.set <- F
                    info$spikegenes <- c() # empty vector
                  }
                else
                                        # add line (trimmed for trailing white space) to gene list
                  {
                    info$spikegenes <- rbind(info$spikegenes,
                                         split.line('fluffypuffy@@###thisoccursnever!',
                                                    sub('[[:space:]]+$', '', curr.str)))
                  }
              }            
          }     
        n <- n+1
      }
    # info$version <- 2
    # use version 1 if version is missing
    if (any(is.na(info$version)))
      info$version <- 1
    if (any(is.na(info$impute.mthd)))
      info$impute.mthd <- "NA"
    if (any(is.na(info$datoutfnam)))
      info$datoutfnam <- "NA"
    if (any(is.na(info$fast.load)))
      info$fast.load <- F
    # now we know whether this is a version 1 or 2 file:
    #if (info$version==1) # then we need to restructure it
    #  {
    #    n.info <- as.list(rep(NA, length(all.inf.nams.v1)))
    #    names(n.info) <- all.inf.nams.v1
    #    n.comp <- length(all.inf.nams.v1)
    #    n <- 1
    #    while(n<=n.comp)
    #      {
    #        n.info[all.inf.nams.v1[n]] <- info[all.inf.nams.v1[n]]
    #        n <- n+1
    #      }
    #    info <- n.info
    #  }
    # check contents of def file
    #print(length(names(info)))
    if(!any(grep('spikegenes', names(info), fixed=T)))
      {
        info$spikegenes <- list()
      }
    is.ok <- fspma.check.deffile(info, deffile=fname, dbg.lognam=log.fname)
    if (!is.ok)
      {
        err.str <- sprintf('Definition file %s shows inconsistencies, check the logfile %s for details.',
                           fname, log.fname)
        stop(err.str)
      }
    # old version: chk.def(info)
    # if moment stabilization is not selected set it false
    if(is.na(info$mom.stab))
      info$mom.stab <- F
    # finally recode spikegenes as list of names and ratios.
    if (!any(is.na(info$spikegenes))&length(info$spikegenes)>0) # only then we may change it
      info$spikegenes <- list(names=info$spikegenes[,1],
                              t.rt=as.numeric(info$spikegenes[,2])/as.numeric(info$spikegenes[,3]))
    else
      info$spikegenes <- list(names=c(), t.rt=c())
    # and rearange such that innermost effects are fixed.
    info <- effect.sort(info)
    info$ngenes <- length(info$geneids)
    info
  }

chk.hdr <- function(sel.hdr, colnams)
  # compare colnams from definition file with
  # the header name at the corresponding position in the
  # file.
  {
    n.data.len <- length(sel.hdr)
    n.chk.len <- length(colnams)
    hdr.ok <- (n.data.len==n.chk.len)
    n <- 1
    while(hdr.ok && n <= n.chk.len)
      {
        hdr.ok <- hdr.ok && sel.hdr[n]==colnams[n]
        n <- n+1
      }
    hdr.ok
  }

parse.colnams <- function(thisline, col.nams)
  {
    # print(thisline)
    strs <- strsplit(thisline, '\t', fixed=T)
    ncols <- length(strs[[1]])
    to.ids <- c(1:ncols)
    nlen <- length(col.nams)
    n <- 1
    ok <- T
    col.nums <- c()
    while(n <= nlen & ok)
      {
        if(any(l.dx <- col.nams[n] == strs[[1]]))
          {
            n.dx <- to.ids[l.dx]
            # print(strs[[1]][n.dx])
            col.nums <- c(col.nums, n.dx)
          }
        else
          {
            # cat(col.nams[n], ' not found\n')
            ok <- F
          }
        n <- n+1
      }
    list(ok=ok, col.nums=col.nums)
  }

get.hdr <- function(col.nams, fnam)
  # searches for header line and returns column numbers in the same order
  # as the names appear in col.nams raises an error if it can't find a line
  # that matches all col.nams after parsing the first 1000 lines
  # returns also the number of lines to skip including the column header to
  # get to the first datum in the file.
  {
    buff <- scan(file=fnam, sep='\n', what='', nlines=50, quiet=T, quote="'\"")
    #buff <- readLines(fnam, n=50)
    nlen <- length(buff)
    #print(buff)
    #print(nlen)
    n <- 0
    go.on <- T
    while(n<nlen & go.on)
      {
        n <- n+1
        ret <- parse.colnams(buff[n], col.nams)
        go.on <- !ret$ok
      }
    if (go.on) # try without quotes
      {
        buff <- scan(file=fnam, sep='\n', what='', nlines=50, quiet=T, quote="")        
        nlen <- length(buff)
        n <- 0
        go.on <- T
        while(n<nlen & go.on)
          {
            n <- n+1
            ret <- parse.colnams(buff[n], col.nams)
            go.on <- !ret$ok
          }
      }
    list(col.nums = ret$col.nums, found = ret$ok, skip = n, line=buff[n])
  }

read.tab.delim <- function (fname, colnams, geneids, spikegenes, minrep, log.fname=fspma.logfile, have.sort.dx=NULL)
  # fname:   file name to read from
  # colnams: header names expected in that column (check consistency of files)
  # geneids: unique list of genes used to read those genes from the slide and order w.r.t them
  # spikegenes: unique list of spike genes to be loaded from the file.
  # minrep:  minimum number of rplicates on all genes and slides
  {    
    #hdr <- read.delim(fname, skip=hdr.len, as.is=T, header=F, nrows=1)
    #hdr <- as.matrix(hdr)
    #hdr <- hdr[,columns]
    #print(hdr)
    #print(colnams)
    #print(columns)
    #if (!chk.hdr(hdr, colnams))
    #  stop(paste('Column name mismatch between definition file and ',
    #             fname, sep='', collapse=''))

    # colnams <- colnams[colnams!='NA']
    hdr <- get.hdr(colnams, fname)
    if (!hdr$found)
      {
        err.strg <- paste(paste('Column names from definition file not found in ',
                                fname, sep='', collapse=''), "\n", sep="", collapse="")
        stream.log(err.strg, fname=log.fname)
        stop(sprintf('See %s for further information.', log.fname))
      }
    skip.lines <- hdr$skip
    columns <- hdr$col.nums
    #print(skip.lines)
    #print(columns)
    tmpframe <- read.delim(fname, skip=skip.lines, as.is=T, header=F)
    tmpframe <- as.matrix(tmpframe)
    tmpframe <- tmpframe[,columns]
    # print(tmpframe)
    # print(dim(tmpframe))
    gene.col <- 1 # genes are in the first column if the reduced frame
    # added to resolve problem with NA entries in gene names...
    nagenes <- is.na(tmpframe[, gene.col])
    tmpframe[nagenes, gene.col] <- "NA"
    # now we parse and split tmpframe    
    len.unique <- length(geneids)
    # print(c(fname, len.unique, 'genes'))
    if (is.null(have.sort.dx))
      {
        sort.dx <- NULL
        genes.ok <- rep(F, len.unique)
        all.dx <- c(1:nrow(tmpframe))
        for (n in 1:len.unique)
          {
                                        # get all hits for this gene
            # no good! produces partial matches this.dx <- grep(geneids[n], tmpframe[, gene.col], fixed=T)
            this.dx <- all.dx[geneids[n]==tmpframe[,gene.col]]
            genes.ok[n] <- length(this.dx)>=minrep
                                        # and take only the first minrep ones
            if (genes.ok[n]) # if enough genes found we add them.
              {
                this.dx <- this.dx[1:minrep]
                dim(this.dx) <- c(1, minrep)
                sort.dx <- rbind(sort.dx, this.dx)
              }
            #else
            #  print(sprintf('%s not found', geneids[n]))
          }
      }
    else
      {
        sort.dx <- have.sort.dx
        genes.ok <- rep(T, len.unique)
      }
    # now we get spike genes Note: the names are needed outside!
    if (length(spikegenes)>0)
      {
        f.gnams <- array(tmpframe[, gene.col], c(length(tmpframe[, gene.col]),1))
        s.gnams <- array(spikegenes, c(1, length(spikegenes)))
        comp <- f.gnams[,rep(1, ncol(s.gnams))]==s.gnams[rep(1, nrow(f.gnams)),]
        dim(comp) <- c(nrow(f.gnams), ncol(s.gnams))
        spike.rows <- apply(comp, 1, any)
        # cat('rows in dataframe from file ',nrow(f.gnams),'\n')
        # cat('number of spike genes found in this file ', sum(spike.rows), '\n')
        spike.frame <- tmpframe[spike.rows,]
      }
    else
      spike.frame <- NULL
    # next: remove geneid, which is not necessary
    tmpframe <- tmpframe[,2:length(columns)]
    if(is.null(dim(tmpframe)))
      dim(tmpframe) <- c(length(tmpframe),1)
    #print(dim(tmpframe))
    #tmp.dim <- dim(tmpframe)
    #dim(tmpframe) <- tmp.dim
    #print(dim(tmpframe))
    # print(sort.dx)
    # now we reorder the rows of dataframe such that each set of replicates
    # is a seperate entry in a list
    # on each other.
    comb.matr <- NULL
    n <- 1
    while (n <= minrep)
      {
        if (minrep>1)
          tmp.mtr <- tmpframe[sort.dx[,n],]
        else
          tmp.mtr <- tmpframe[sort.dx,]
        dim(tmp.mtr) <- c(nrow(sort.dx), ncol(tmpframe))
        comb.matr <- rbind(comb.matr, tmp.mtr)
        n <- n+1
      }
    #print('end reading')
    list(comb.matr=comb.matr, genes.ok=genes.ok, spike.frame=spike.frame, sort.dx=sort.dx)
  }


# Helper for building RG structure: convert info$dat.file.rd[n]$expdx into
# index in a matrix (where data from file is written).

effcnts2idx <- function(effcnts, effdims, ngenes, nrep=1)
  {
    # effdims contains all efefct dimensions WITHOUT replicates on slides
    # effcnts contains the effect id's of the current file.
    # ngenes is the number of genes that go into the analysis.
    nlen <- length(effdims)
    n <- nlen
    dx <- 0
    fkt <- 1
    while (n>0)
      {
        dx <- dx+fkt*ngenes*nrep*(effcnts[n]-1)
        fkt <- fkt*effdims[n]
        n <- n-1
      }
    dx <- dx+c(1:(ngenes*nrep))
    #print(dx)
    dx
  }

#helper function: generate default values for RG entries if they are not available
# move.2.log indicates whether we move the RG to log space. In that instance the neutral
# element is 1 (since we take log ratios) otherwise it is 0 since we take differences.
rg.defaults <- function(name, dimensions, move.2.log)
  {
    if (move.2.log==T)
      ret <- switch(name,
                    'x'=, 'y'=, 'Rb'=, 'Gb'=, 'Flg'=array(rep(0, prod(dimensions)), dim=dimensions),
                    'gridno'=, 'slideno'=, 'G'= array(rep(1, prod(dimensions)), dim=dimensions))
    else
      ret <- switch(name,
                    'x'=, 'y'=, 'Rb'=, 'Gb'=, 'Flg'=, 'G'=array(rep(0, prod(dimensions)), dim=dimensions),
                    'gridno'=, 'slideno'= array(rep(1, prod(dimensions)), dim=dimensions))
    ret
  }

#helper: change type
spike.type.cast <- function(vals, name, is.num.flag=T)
  {
    dims <- dim(vals)
    if (is.num.flag)
      vals <- switch(name,
                     'x'=, 'y'=, 'R'=, 'Rb'=, 'G'=, 'Gb'=, 'gridno'=, 'slideno'=, 'Flg' = as.numeric(vals),
                     'genes'=as.character(vals))
    else
      vals <- switch(name,
                     'x'=, 'y'=, 'R'=, 'Rb'=, 'G'=, 'Gb'=, 'gridno'=, 'slideno' = as.numeric(vals),
                     'Flg'=, 'genes'=as.character(vals))
    dim(vals) <- dims
    vals
  }

# helper: build the RG$normspike frame from the frame that was read from file.
# (this is to.nams) RG.listnams <- c('gene', 'R', 'Rb', 'G', 'Gb', 'Flg', 'x', 'y', 'gridno', 'slideno')
spike2normframe <- function (spike.frame, normspikes, spike.ifo, is.avail, to.nams, slideno,
                             is.num.flag, move.2.log)
  {
    #print(spike.frame)
    #print(is.avail)
    #print(to.nams)
    # first prepare the slide information to be appended
    #at('spikes slide: ', slideno,'\n')
    #print(ncol(spike.frame))
    #print(length(spike.frame[,1]))
    if(!is.null(spike.frame))
      {
        #print('rubbish')
        dat.dims <- c(length(spike.frame[,1]), 1)
        n.frame <- NULL
        n.cols <- length(is.avail)
        n <- 1
        dat.cntr <- 1
        while (n <=n.cols)
          {
            #cat('column ',n, ' ', to.nams[n], ' available ', is.avail[n], '\n')
            if (is.null(n.frame))
              {
                if(is.avail[n])
                  {
                    n.frame <- data.frame(I(spike.type.cast(spike.frame[,dat.cntr], to.nams[n], is.num.flag)))
                    dat.cntr <- dat.cntr+1
                  }
                else
                  {
                    tmp.data <- rg.defaults(to.nams[n], dat.dims, move.2.log)
                    dim(tmp.data) <- NULL
                    n.frame <- data.frame(I(spike.type.cast(tmp.data, to.nams[n], is.num.flag)))
                  }
              }
            else
              {
                if(is.avail[n])
                  {
                    n.frame <- cbind(n.frame, data.frame(I(spike.type.cast(spike.frame[,dat.cntr],
                                                                           to.nams[n], is.num.flag))))
                    dat.cntr <- dat.cntr+1
                  }
                else
                  {
                    tmp.data <- rg.defaults(to.nams[n], dat.dims, move.2.log)
                    dim(tmp.data) <- NULL
                    n.frame <- cbind(n.frame, data.frame(I(spike.type.cast(tmp.data, to.nams[n], is.num.flag))))
                  }
              }
            spike.type.cast(n.frame[,n], to.nams[n], is.num.flag)
            n <- n+1
          }
        # now we may append the slideno
        slideno <- rep(slideno, length(spike.frame[,1]))
        n.frame <- cbind(n.frame, data.frame(slideno))
        # last part: append the true ratio of spike concentrations
        n.genes <- length(spike.ifo$names)
        t.rt <- rep(1, n.genes)
        n <- 1
        while(n <= n.genes)
          {
            t.rt[n.frame[,1]==spike.ifo$names[n]] <- spike.ifo$t.rt[n]
            n <- n+1
          }
        n.frame <- cbind(n.frame, data.frame(as.numeric(t.rt)))
        # cat('cols normspike ', ncol(normspikes))
        fr.nam <- c(to.nams, 't.rt')
        names(n.frame) <- fr.nam
        # and finally row bind with normspikes
        normspikes <- rbind(normspikes, n.frame)
        # print(fr.nam)
        names(normspikes) <- fr.nam
      }
    normspikes
  }
# helper: normalize X and Y positions accross all slides
# this avopids that very large values lead to numerical difficulties in
# the normalization procedure.
norm.space.lin <- function(RG)
  {
    x.mn <- 0
    sdev.x <- 1
    y.mn <- 0
    sdev.y <- 1
    if(!any(!is.finite(RG$x)))
      {
        x.mn <- mean(RG$x)
        RG$x <- RG$x - x.mn
        sdev.x <- sd(as.vector(RG$x))
        if(sdev.x>fspma.eps)
          RG$x <- RG$x/sdev.x
      }
    else
      {
        idx.nonfinite <- !is.finite(RG$x)
        cat('|',RG$x[idx.nonfinite],'|', RG$y[idx.nonfinite], '|\n')
        stop('Warning: Some Spot X-Positions are not finite (character strings?)!\n')
      }
    if(!any(!is.finite(RG$y)))
      {
        y.mn <- mean(RG$y)
        RG$y <- RG$y - y.mn
        sdev.y <- sd(as.vector(RG$y))
        if(sdev.y>fspma.eps)
          RG$y <- RG$y/sdev.y
      }
    else
      {
        idx.nonfinite <- !is.finite(RG$y)
        cat('|',RG$x[idx.nonfinite],'|', RG$y[idx.nonfinite], '|\n')
        stop('Warning: Some Spot Y-Positions are not finite (character strings?)!\n')
      }
    if(length(RG$normspike)>0) # then we also have spike data 
      {
        if(!any(!is.finite(RG$normspike$x)))
          {
            # do the same to the spike positions
            RG$normspike$x <- RG$normspike$x-x.mn
            if(sdev.x>fspma.eps)
              RG$normspike$x <- RG$normspike$x/sdev.x
          }
        else
          {
            idx.nonfinite <- !is.finite(RG$normspike$x)
            cat('|', RG$normspike$x[idx.nonfinite], '|', RG$normspike$y[idx.nonfinite], '|\n')
            stop('Warning: Some Spike X-Positions ar not finite (character strings?)!\n')
          }
        if(!any(!is.finite(RG$normspike$y)))
          {
            # do the same to the spike positions
            RG$normspike$y <- RG$normspike$y-y.mn
            if(sdev.y>fspma.eps)
              RG$normspike$y <- RG$normspike$y/sdev.y
          }
        else
          {
            idx.nonfinite <- !is.finite(RG$normspike$y)
            cat('|', RG$normspike$x[idx.nonfinite], '|', RG$normspike$y[idx.nonfinite], '|\n')
            stop('Warning: Some Spike Y-Positions ar not finite (character strings?)!\n')
          }
      }
    RG
  }

# internal use only:
stripgenes <- function(info, ofile.ext='tsv', log.fname=fspma.logfile)
  # removes unwanted genes and otherwise writes file as is. changes extension
  # of output file.
  {
    # first we determine the actually specified column names.
    colnams <- c(info$genecolnam, info$rcolnam, info$rbcolnam, info$gcolnam,
                 info$gbcolnam, info$flagcolnam, info$xcolnam, info$ycolnam,
                 info$gridnocolnam)
    isavail <- colnams!='NA' & !is.na(colnams)
    colnams <- colnams[isavail]
    # now we loop through dat.file.rd to get the reduced files
    nfiles <- length(info$dat.file.rd)
    f.ctr <- 1
    while(f.ctr <= nfiles)
      {
        # prepare names
        infile <- info$dat.file.rd[[f.ctr]]$fnam
        nchr <- strlen(infile)
        #print(nchr)
        if(substr(infile,nchr-3, nchr-3)=='.')
          ofile <- paste(substr(infile,1,(nchr-3)), ofile.ext, sep="", collapse="")
        else
          ofile <- paste(infile, ofile.ext, sep=".", collapse="")
        strg <- sprintf('Converting %s to %s\n', infile, ofile);
        stream.log(strg, fname=log.fname)
        # get header
        hdr <- get.hdr(colnams, infile)
        if (!hdr$found)
          stop('File specification error in info / definition file - extraction of genes fails.')
        # indices of columns:
        coldx <- hdr$col.nums
        tmpframe <- read.delim(infile, skip=hdr$skip, as.is=T, header=T)
        tmpframe <- tmpframe[,coldx]
#        names(tmpframe) <- colnams[coldx]
        #print(tmpframe)
        #print(dim(tmpframe))
        #print(nrow(tmpframe))
        #print(ncol(tmpframe))
        #print(tmpframe[c(1:5),])
        gene.col <- 1 # genes are in the first column if the reduced frame
        #print(gene.col)
        #print(tmpframe[,gene.col])
        genesinfile <- tmpframe[,gene.col]
        # now we parse and split tmpframe
        geneids <- unique(info$geneids)
        len.unique <- length(geneids)
        # print(c(fname, len.unique, 'genes'))
        sort.dx <- NULL
        for (n in 1:len.unique)
          {
            # get all hits for this gene
            this.dx <- grep(geneids[n], genesinfile, fixed=T)
            dim(this.dx) <- c(length(this.dx), 1)
            sort.dx <- rbind(sort.dx, this.dx)
          }
        tmpframe <- tmpframe[sort.dx,]
        # finally we are in position to write that dataset to a tab delimited file:
        write.table(tmpframe, file = ofile, row.names = FALSE, sep="\t")
        f.ctr <- f.ctr + 1
      }
  }

# Next function: Build RG structure from the def-file description and all files that
# should be available.
tabdel2rg <- function(info, log.fname=fspma.logfile)
# function reads in data according to specification in def-file.
  {
    neffects <- length(info$effects)
    dx.effdims <- info$effects[1:(neffects-1)]
    # determine the situation:
    # genecol and rcol must be available here.
    #cols <- c(info$genecol, info$rcol, info$rbcol, info$gcol,
    #          info$gbcol, info$flagcol, info$xcol, info$ycol, info$gridnocol)
    colnams <- c(info$genecolnam, info$rcolnam, info$rbcolnam, info$gcolnam,
                 info$gbcolnam, info$flagcolnam, info$xcolnam, info$ycolnam,
                 info$gridnocolnam)
    # print(colnams)
    #swp.cols <- c(info$genecol, info$gcol, info$gbcol, info$rcol,
    #              info$rbcol, info$flagcol, info$xcol, info$ycol, info$gridnocol)
    swp.colnams <- c(info$genecolnam, info$gcolnam, info$gbcolnam, info$rcolnam,
                     info$rbcolnam, info$flagcolnam, info$xcolnam, info$ycolnam,
                     info$gridnocolnam)
    # names in RG structure for data. Flg does not exist and
    # is generated in addition to allow to work with flags (extension)
    RG.listnams <- c('genes', 'R', 'Rb', 'G', 'Gb', 'Flg', 'x', 'y', 'gridno', 'slideno')
    # first we generate two vectors: one that indicates existing entries
    # and one that indicates non existing entries.
    # cat('length cols ', length(cols),' \n')
    isavail <- colnams!='NA' & !is.na(colnams) # !is.na(cols)
    isnan <- !isavail  # is.na(cols)
    # rdcols <- cols[isavail]
    n.cols <- sum(isavail)
    # old - remove 1 for genes
    # n.d.cols <- length(rdcols)-1
    # new: we have one column more for the slide no which is not part of the columns.
    n.d.cols <- n.cols
    rdcolnams <- colnams[isavail]
    # swp.rdcols <- swp.cols[isavail]
    swp.rdcolnams <- swp.colnams[isavail]
    nfiles <- length(info$dat.file.rd)
    # now check whether there is a color swap and in this case whether the symetry requirement
    # is met.
    is.swap <- F
    n <- 1
    #cat('nf ', nfiles, ' n ', n,' swp ', is.swap) 
    while (n <= nfiles & !is.swap)
      {
        is.swap <- info$dat.file.rd[[n]]$doswp
        n <- n+1
        #cat('nf ', nfiles, ' n ', n,' swp ', is.swap) 
      }
    proc.dual.swap <- F
    if (is.swap & any(swp.rdcolnams=='NA'))
      {
        # we first make sure that there is no G, Rb amd Gb channel
        if(info$gcolnam!='NA' | info$rbcolnam != 'NA' | info$gbcolnam !='NA')
          {
            strg <- 'Error with colour swap - symmetry broken (R requires G channel and Rb requires Gb channel).';
            #strg <- sprintf('Reading File: %s \n', info$dat.file.rd[[n]]$fnam)
            stream.log(strg, fname=log.fname)
            stop(' ')
          }
        proc.dual.swap <- T
        swp.rdcolnams <- rdcolnams 
        strg <- 'Processing assumes color swap for preprocessed dual color arrays.\n';
        # strg <- sprintf('Reading File: %s \n', info$dat.file.rd[[n]]$fnam)
        stream.log(strg, fname=log.fname)
      }
    # construct data store for all data.
    n.rep <- info$effects[neffects]
    n.rows <- prod(info$effects)*info$ngenes
    is.data.set <- rep(F, n.rows)
    n.stores <- n.rows*n.d.cols
    tmp.store <- rep(0, n.stores)
    dim(tmp.store) <- c(n.rows, n.d.cols)
    d.store <- tmp.store
    dim(d.store) <- dim(tmp.store)
    # construct store for effects.
    n.effrows <- prod(info$effects)
    is.eff.set <- rep(F, n.effrows)
    #print(info$effects)
    #slrep <- info$effects[length(info$effects)]
    eff.store <- array(rep(NA, n.effrows*(length(info$effects))), c(n.effrows, length(info$effects)))
    #print('eff store')
    #print(dim(eff.store))
    #print(length(eff.store))
    # construct flags for handling processed channel color swap.
    swp.flgs <- rep(F, n.rows)
#    print('length swp.flgs')
#    print(n.rows)
    normspikes <- data.frame()
    # read all files data
    have.gene.dx <- NULL
    n <- 1
    while (n <= nfiles)
      {
        # first get data from file
        # print(info$dat.file.rd[[n]]$fnam)
        strg <- sprintf('Reading File: %s \n', info$dat.file.rd[[n]]$fnam)
        stream.log(strg, fname=log.fname)
        if (info$dat.file.rd[[n]]$doswp==T) # we have swapped colors in this file
          ret <- read.tab.delim(info$dat.file.rd[[n]]$fnam, swp.rdcolnams, info$geneids,
                                info$spikegenes$names, n.rep, log.fname, have.gene.dx)
        else
          ret <- read.tab.delim(info$dat.file.rd[[n]]$fnam, rdcolnams, info$geneids,
                                info$spikegenes$names, n.rep, log.fname, have.gene.dx)
        if (n==1 & info$fast.load) # then all subsequent files have the same gene positions.
          have.gene.dx <- ret$sort.dx
        this.data <- ret$comb.matr
        # print(dim(this.data))
        # add slide number here:
        this.slide <- data.frame(rep(n, nrow(this.data)))
        # names(this.slide) <- 'slideno'
        this.data <- cbind(this.data, this.slide)
        if (any(!ret$genes.ok))
          {
            strg <- sprintf('The following genes in file $s have not enough replicates on slide:\n',
                info$dat.file.rd[[n]]$fnam)
            stream.log(strg, fname=log.fname)
            strg <- paste(paste("", info$geneids[!ret$genes.ok], sep="\n", collapse=""), "\n", sep="", collapse="")
            stream.log(strg, fname=log.fname)
            strg <- sprintf("There are %3.0f genes with missing replicates.\n", sum(!ret$genes.ok)) 
            pos <- c(1:length(ret$genes.ok))
            strg <- paste(strg, paste("", pos[!ret$genes.ok], sep="\n", collapse=""), sep="", collapse="")
            strg <- paste(strg, "\n", sep="", collapse="")
            stream.log(strg, fname=log.fname)
            stop(sprintf('Correct the definition file, see log file, %s, for details!', log.fname))
          }
        # this.data has now the same number of entries as we have stores in
        # d.store (one for each replicates)
        # we get the column position of the contents of this.data in d.store
        #print('calc dx')
        #print(info$dat.file.rd[[n]])
        #print(dx.effdims)
        #print(info$ngenes)
        #print(n.rep)
        #print('get dx')
        dx <- effcnts2idx(info$dat.file.rd[[n]]$expdx, dx.effdims, info$ngenes, n.rep)
        # we have data for those indices
        is.data.set[dx] <- T
        # if we have a swap we set the index in swp.flgs TRUE
        if (info$dat.file.rd[[n]]$doswp==T)
          swp.flgs[dx] <- T
        #print('have dx')
        #print(dim(d.store))
        #print(dx)
        #print(dim(this.data))
        # and write all replicates into that position in store
        d.store[dx,] <- as.matrix(this.data)
        #print('writing in d.store done')
        #print(dim(d.store))
        # next consider the effects 
        eff.dx <- effcnts2idx(info$dat.file.rd[[n]]$expdx, dx.effdims, 1, n.rep)
        these.vals <- t(array(rep(info$dat.file.rd[[n]]$expdx, n.rep),
                              c(length(info$dat.file.rd[[n]]$expdx), n.rep)))
        these.vals <- cbind(these.vals, c(1:n.rep))
        #print(these.vals)
        #print(eff.store[eff.dx,])
        eff.store[eff.dx,] <- these.vals
        is.eff.set[eff.dx] <- T
        #print('writing eff.store done')
        # now we consider the spike frame (ret$spike.frame) as readfrom the file
        normspikes <- spike2normframe(ret$spike.frame, normspikes, info$spikegenes,
                                      isavail,  RG.listnams, slideno=n, is.numeric(info$flag.nack),
                                      info$dologarray)
        #print(dim(d.store))
        #print('have file one')
        n <- n+1
      }
    # to allow for unbalanced designs, we will as a next step reduce both datastores by all
    # effects that were not initialised during the setup
    d.store <- d.store[is.data.set,]
    eff.store <- eff.store[is.eff.set,]
    # since we have added the slide number to the information that is read from
    # the file, we increase the length of isavail by one true entry:
    isavail <- c(isavail, T)
    # now all files have been read and we refromat the data to the RG structure
    data.dim <- c(length(info$geneids), sum(is.eff.set))
    RG <- list(genes=info$geneids, eff.store=eff.store)
    n.exps <- length(info$effects)
    n <- 1
    experiments <- list()
    while(n <= n.exps)
      {
        add.exp <- list(1:info$effects[n])
        names(add.exp) <- info$effnams[n]
        experiments <- c(experiments, add.exp)
        n <- n+1
      }
    RG <- c(RG, list(experiments=experiments))
    list.len <- length(RG.listnams)
    n <- 2
    dt.col <- 1
    # go through list and construct the respective part of the structure
    while(n<=list.len)  
      {
        if(isavail[n]) # we have the data in the store
          {
            # now we have a special case for Flags! (they might be numeric or string)
            if ('Flg'==RG.listnams[n])
              {
                if (is.character(info$flag.nack)) # here we take the flag values as is
                  {
                    tmp.dat <- d.store[,dt.col]
                    dim(tmp.dat) <- data.dim
                  }
                else # here we convert them to numerical values
                  {
                    tmp.dat <- as.numeric(d.store[,dt.col])
                    dim(tmp.dat) <- data.dim
                  }
              }
            else # we have channel values or other numeric information
              {
                tmp.dat <- as.numeric(d.store[,dt.col])
                dim(tmp.dat) <- data.dim
              }
            add.dat <- list(tmp=tmp.dat)
            names(add.dat) <- RG.listnams[n]
            RG <- c(RG, add.dat)
            dt.col <- dt.col+1
          }
        else
          {
            add.dat <- list(tmp=rg.defaults(RG.listnams[n], data.dim, info$dologarray))
            names(add.dat) <- RG.listnams[n]
            RG <- c(RG, add.dat)
          }
        n <- n+1
      }
    # finally we treat the color swap in case of processed dual color files that come as log ratios
    if (proc.dual.swap & any(swp.flgs))
      {
        #print('resolve color swap')
        #print(length(swp.flgs))
        #print(prod(dim(RG$R)))
        # channel values
        tmp <- RG$R[swp.flgs]
        RG$R[swp.flgs] <- RG$G[swp.flgs]
        RG$G[swp.flgs] <- tmp
        # background values
        tmp <- RG$Rb[swp.flgs]
        RG$Rb[swp.flgs] <- RG$Gb[swp.flgs]
        RG$Gb[swp.flgs] <- tmp        
      }
    # attach normspikes as list entry normspike
    RG <- c(RG, list(normspike=normspikes))
    # and normalize the spatial information
    #RG <- norm.space.lin(RG) -> moved to fspma.wrapper!
    RG
  }

# now we generate the anova equations from info. 
# the following is a new function that generates all model formulae
# and additional information necessary to apply YASMA to longitudinal problems
gen.yasma.eqns <- function(terms, randeffs, rank.term=1)
  # gen.yasma.eqns generates default model equations and
  # additional information that I use to calculate
  # test statstics. terms does not include gene names.
  # however all other fixed and random factors in the
  # experiment. randeffs is a vector of length equal to terms
  # indicating for every term, whether it is a random effect (T) or
  # a fixed effect (F). rank.term is the index of the term that
  # defines the ranking we are interested in. (Typically this will be
  # a gene:time, a gene:species or a gene:type interaction.)
  # and the index will thus be 1.
  {
    MSav.dcol <- 0 # init if 0 we use MSE and dfe for comparing!
    MSav.ncol <- rank.term
    nt <- length(terms)
    eqn <- '~G' # defalt term for gene
    eq.rand <- '~'
    eq.tsteff <- '~'
    eq.run <- 'G'
    eq.avrnk <- '~'
    eq.avrun <- ''
    rand.sel <- F
    for(n in 1:nt)
      {
        eq.run <- paste(eq.run, terms[n], sep=':')
        eqn <- paste(eqn, eq.run, sep='+')
        if (n==rank.term)
          {
            eq.avrun <- paste(eq.avrun, terms[n], sep='')
            eq.avrnk <- paste(eq.avrnk, eq.avrun, sep='')
          }
        if (n > rank.term)
          {
            eq.avrun <- paste(eq.avrun, terms[n], sep=':')
            eq.avrnk <- paste(eq.avrnk, eq.avrun, sep='+')
          }
        if (rand.sel & randeffs[n])
          {
            rand.sel <- F
            eq.rand <- paste(eq.rand, eq.run, sep='')
            MSav.dcol <- n
          }
        if (n==rank.term)
          {
            rand.sel <- T
            eq.tsteff <- paste(eq.tsteff, eq.run, sep='')
          }
      }
    if (eq.rand=='~')
      eq.rand <- '~ 1'
#    print(eqn)
#    print(eq.rand)
#    print(eq.tsteff)
#    print(eq.avrnk)
#    print(c( MSav.ncol, MSav.dcol))
    if (rank.term >= 1)
      ret <- list(eqn=as.formula(eqn), eq.rand=as.formula(eq.rand),
                  eq.eff=as.formula(eq.tsteff), eq.avrnk=as.formula(eq.avrnk),
                  MSav.ncol=MSav.ncol, MSav.dcol=MSav.dcol)
    else # we only need the model equation in this case! The other equations are initialised as defaults
      ret <- list(eqn=as.formula(eqn), eq.rand=as.formula(eq.rand),
                  eq.eff=as.formula(eq.rand), eq.avrnk=as.formula(eq.rand),
                  MSav.ncol=MSav.ncol, MSav.dcol=MSav.dcol)
  }

info.2.yasmaeqns <- function(info)
  # get yasma equations from info structure
  {
    # info$effnams: names of effects for equations and data structure
    # info$effects: number of effects and repetitions within each
    # info$randeffs: boolean info about random effects
    # info$rank.term: rank term used by anova based ranking
    # important: we need to remove on slide replications from the
    # equations. We thus remove the last component from effnams and
    # randeffs.
    n.effs <- length(info$effnams)
#    if(info$effects[n.effs]>1) # we have multiple replicates on slide
#      {
        effnams <- info$effnams[1:(n.effs-1)]
        randeffs <- info$randeffs[1:(n.effs-1)]
#      }
#    else      # we have only one replicate and thus need to allow the last but one for the residuals
#      {
#        effnams <- info$effnams[1:(n.effs-2)]
#        randeffs <- info$randeffs[1:(n.effs-2)]        
#      }
    gen.yasma.eqns(effnams, randeffs, info$rank.term)
  }

# wrapp anova
do.anova <- function(model.info, array.data)
  {
    #cat('do anova \n')
    bfaov(model.info$eqn, array.data)
    #cat('anova done \n')
  }

# write anova table to file
# aov.tab.2.file <- function(av, file)
#  {
#    sink(file)
#    anova(av)
#    sink()
#  }

patch.reml.R19 <- function(info, model.info, rml=NULL)
  # reml coords patch that is necessary (at least for R 1.9)
  # since the original code does not get the names of the variance
  # components right, if there is only a single occurence of a random
  # effect.
  {
    require(nlme)
    # get random effects
    rand.effnams <- info$effnams[info$randeffs]
    n.randeffs <- length(rand.effnams)
    if (n.randeffs>1) # the last random effect is for the residual (anova output!)
      rand.effnams <- rand.effnams[1:(n.randeffs-1)]
    else              # fixed effects model and thus no other variance components!
      rand.effnams <- NULL
    reml.coords <- list()
    if(!is.null(rand.effnams)) # then patch their coords
      {
        eq <- model.info$eqn
        # split formula
        sf.p <- splitFormula(eq, '+')
        n.len <- length(sf.p)
        n <- 1
        while (n <= n.len)
          {
            tmp <- splitFormula(sf.p[[n]], ':')
            eff.nams <- NULL
            for (i in 1:length(tmp))
              {
                eff.nams <- c(eff.nams, attr(terms(tmp[[i]]), 'term.labels'))
              }
            eff.dims <- match(eff.nams, rand.effnams, nomatch=0)!=0
            if (any(eff.dims))
              reml.coords <- c(reml.coords, list(eff.nams))
            n <- n+1
          }
      }
    if (length(reml.coords)==1) # only then we need the patch! 
      rml$coords <- c(reml.coords, list('res'))
    rml
  }

# av.ou <- over

# write anova table and variance components to file
aov.tab.varcomp.2.file <- function(av, info, model.info, file=NULL)
  {
    if (is.null(file)) # then init file name from info
      file <- info$aovfname
    if (! any(file=='NA')) # NA is again used as not applicable
      {
        # get random effects names from info structure
        rand.effnams <- info$effnams[info$randeffs]
        n.randeffs <- length(rand.effnams)
        if (n.randeffs>1) # the last random effect is for the residual (anova output!)
          rand.effnams <- rand.effnams[1:(n.randeffs-1)]
        else              # fixed effects model and thus no other variance components!
          rand.effnams <- NULL
        if(!is.null(rand.effnams)) # then do the reml calculations
           {
             # call true reml
             rml <- reml.bfaov(av, rand.effnams, model.info$eqn, anova.only=F)
             rml <- patch.reml.R19(info, model.info, rml)
           }
        # we first write the ANOVA result and then variance components into the file
        sink(file)
        cat('ANOVA Table:\n\n')
        anova(av)
        if(!is.null(rand.effnams)) # then we did the reml calculations and write them into the file
          {
            cat('\n\n')
            cat('Variance Components:\n\n')
            # print variance components to sink
            print.reml(rml)
          }
        sink(file=NULL)
      }
  }

# The following implements contrast based ranking within anova which uses ou.int.diff
# and ou.int.sd, together with the evaluation functions ou.int.get and
# ou.int.prnt to tabulate selected effects.

ou.int.diff <- function(av, gt.eff)
  # calculate the mean of the interaction effects we want to check
  # for significant up or down regulation.
  # example call:
  # allmns <- ou.int.diff(av, ~G:time)
  {
    # YASMA stores the anova data internally with dimensions named
    # by the effect names. Hence we get the dimensions to search over by
    if (any(names(av)=='data')) # then we call with an anova object
      a <- av$data
    else # we assume a call with a suitably shaped data object (like in the conversion to average expressions)
      a <- av
    # print(a)
    require(nlme)
    # split formula
    tmp <- splitFormula(gt.eff, ':') 
    # print(tmp)
    eff.nams <- NULL
    # print(length(tmp))
    for (i in 1:length(tmp))
      {
        # print(attr(terms(tmp[[i]]), 'term.labels'))
        eff.nams <- c(eff.nams, attr(terms(tmp[[i]]), 'term.labels'))
      }
    eff.dims <- match(eff.nams, names(dim(a)))
    #print(eff.dims)
    # and a matrix of means taken over all other effects
    eff.mean <- apply(a, eff.dims, mean)
    d.dims <- dim(a)
    dim(eff.mean) <- d.dims[eff.dims]
    #print(eff.mean)
    eff.mean
  }

ou.g.t.se <- function(av, rand.eff, gt.eff)
  # calculate the standard error of the mean gene time interaction effect we want
  # to check for significant up or down regulation and the degrees of freedom of
  # the corresponding mean square (i.e. the "precision" of the variance estimate).
  # the parameter is the anova object as generated by bfaov.
  # rand.eff is the random effect that corresponds to the
  # appropriate mean suqre which gives the std. dev. of the
  # gene:time interatcion. gt.eff is the interaction effect we want to
  # check. (could be G as well!)
  # example call:
  # se <- ou.g.t.se(av, ~G:time:smpl, ~G:time)
  {
    if (!is.null(rand.eff))
      {
        # get mse of random effect (this is necessarily an interaction term!)
        rand.eff <- terms(rand.eff)
        rand.eff <- attr(rand.eff,'term.labels')
        t <- terms(av$formula)
        nams <- t(attr(t,'term.labels'))
        MScol <- grep(rand.eff, nams, fixed=T)
        tmp <- anova.from.comblist(av)
        MS <- tmp$MS[MScol]
        df <- tmp$df[MScol]
      }
    else
      {
        # residual mse
        an.res <- anova(av, ptype='none')
        MS <- an.res$MS
        df <- an.res$df
      }
    # finally get number of objects that contribute to the
    # mean of the fixed effect (interaction) we are interessted in
    # testing.
    tmp <- av$dims
    dim(tmp) <- c(1, length(tmp))
    colnames(tmp) <- names(av$dims)
    n.obj <- sum(tmp)-sum(model.frame(gt.eff, data.frame(tmp)))
    # tmp <- anova.from.comblist(av) ????
    # mean square in MS sse is the squared std. error of mean
    #print(MS)
    #print(n.obj)
    #print(df)
    list(sse=MS/n.obj, df=df)
  }

mysd <- function(x)
  {
    dim(x) <- NULL
    sdev <- sd(x)
    # print(length(sdev))
    sdev
  }

order.4.file <- function(p.vals, is.up, g.nams, all.mns, data.colnams, base.level=1,
                         contrast.name=NULL)
  {
    #cat('base.level:', base.level, '\n')
    p.cols <- ncol(p.vals)
    cc <- 1
    data.ret <- c()
    if (!is.null(contrast.name))
      {
        l.nams <- contrast.name  # name by contrast.name
      }
    else
      {
        if (length(data.colnams)>1)
          l.nams <- data.colnams[-base.level]   # name by all dimensions appart from base level
        else
          l.nams <- data.colnams  # name by base level if a Cy5/Cy3 comparison
      }
    # print(p.cols)
    # print(l.nams)
    if (p.cols != length(l.nams))
      stop('Error contrast names do not match number of p-value columns')
    l.nams <- paste('test', l.nams, sep='.')
    #print(p.cols)
    #print(dim(p.vals))
    #print(dim(is.up))
    #print(length(g.nams))
    #print(dim(all.mns))

    while(cc <= p.cols)
      {
        #print(cc)
        this.p <- p.vals[,cc]
        this.up <- is.up[,cc]
        #print('have this')
        # sort the pvalues
        rwdx <- sort(this.p, index.return=T)$ix
        this.p <- this.p[rwdx]
        this.up <- this.up[rwdx]
        #print('have sort this')
        g.srt <- g.nams[rwdx]
        #print('genes sort')
        if (length(dim(all.mns)) > 2)
          {
            nlen <- length(rwdx)
            ndat <- NULL
            for (n in 1:nlen)
              {
                ndat <- rbind(ndat, all.mns[slice.index(data, 1)==rwdx[n]])
              }
            s.mns <- ndat
          }
        else
          {
            s.mns <- all.mns[rwdx,]
          }
        #print('have means srt')
        dim(s.mns) <- dim(all.mns)
        this.dim <- c(length(this.p),1)
        dim(this.p) <- this.dim
        dim(this.up) <- this.dim
        dim(g.srt) <- this.dim
        # now we add this sorted comparison to the list of all.
        this.data <- data.frame(rwdx, I(g.srt), this.p, this.up, s.mns)
        mn.nams <- data.colnams
        mn.nams <- paste(mn.nams, 'mean', sep='.')
        names(this.data) <- c('dx', 'g.nams', 'p.val', 'is.up', mn.nams)
        if(cc==1)
          {
            data.ret <- list(this.data)
            names(data.ret) <- l.nams[cc]
          }
        else
          {
            add.this <- list(this.data)
            names(add.this) <- l.nams[cc]
            data.ret <- c(data.ret, add.this)
          }
        cc <- cc+1
      }
    #print('bye')
    data.ret
  }

ou.int.get <- function (av, RG, model.info, info, data.colnams, gt.eff, rand.eff=NULL,
                        base.level=1, contrast=NULL, contrast.name=NULL)
# calculate a list of differentially expressed genes, sorted w.r.t.
# min p-value across time (override by sort.sgle=F where we sort w.r.t. the single
# most significant p-values in each row. base.level is the level we conpare
# against. Default: 1 which in mouse.testis is the adult generation.
# av: bfaov object, RG: YASMA RG object (only for gene names)
# gt.eff: model formula for interaction effects we want to check for.
# rand.eff=NULL: assume fixed effects model and base the calculation on
# the residual noise. otherwise: model formula of relevant MSE interaction.
# This depends on the model equation used in the ANOVA and on the assumption
# which of the effects are random.
{
  if (!is.null(contrast))
    {
      if (abs(sum(contrast))>10*fspma.eps)
        stop('Wrong contrast definition')
    }
  # get standard error
  res.se <- ou.g.t.se(av, rand.eff, gt.eff)
  se.diff <- sqrt(2*res.se$sse)
  df <- res.se$df
  #print(se.diff)
  #print(df)
  # get table of means
  #print('start means')
  allmns <- ou.int.diff(av, gt.eff)
  #print(allmns)
  #print(dim(allmns))
  #print('have means')
  # we are now ready to calculate the p-values for all pairwise comparisons.
  # first: get mean difference between base.level and all other levels.
  if (any(is.null(contrast)))
    {
      if(length(allmns[,-base.level])>0) # then we have a comparison among levels
        mn.diff <- allmns[,-base.level] - allmns[, base.level]
      else # we look for significant differential expression between cy3 and cy5
        mn.diff <- allmns[, base.level]
      # added: get up or downregulation w.r.t. base level
      is.up <- mn.diff>0
      mn.diff <- abs(mn.diff)
      mn.diff <- mn.diff/se.diff
      p.dims <- dim(mn.diff)
      # calculate the p-values for a t-distribution with df degrees of freedom.
      p.vals <-  2*pt(mn.diff, df=df, lower.tail=F) # note the 2 ! this is a two sided problem
    }
  else
    { # new contrast based p-values
      nmns <- nrow(allmns)
      multcols <- rep(1, nmns)
      dim(contrast) <- c(1, length(contrast))
      # print(dim(contrast))
      #print(dim(allmns))
      mn.diff <- apply(allmns*contrast[multcols,], 1, sum)
      if (is.null(dim(mn.diff)))
        {
          dim(mn.diff) <- c(nmns, 1)
        }
      #print(mn.diff)
      is.up <- mn.diff > 0
      mn.diff <- abs(mn.diff)
      se.contfact <- sum(contrast^2)
      se.diff <- sqrt(se.contfact*res.se$sse)
      mn.diff <- mn.diff/se.diff
      p.dims <- dim(mn.diff)
      #print(p.dims)
      # calculate the p-values for a t-distribution with df degrees of freedom.
      p.vals <- 2*pt(mn.diff, df=df, lower.tail=F) # note the 2 ! this is a tw
    }
  dim(p.vals) <- p.dims
  dim(is.up) <- p.dims
  if (is.null(p.dims))
    {
      n.p <- length(p.vals)
      dim(p.vals) <- c(n.p, 1)
      dim(is.up) <- c(n.p, 1)
    }
  # number of comparisons equals the number of entries in the p.vals array
  # (modification for contrast based vesrion)
  n.comps <- prod(dim(p.vals))
  g.nams <- RG$gene
  g.up <- is.up
  dim(g.nams) <- c(length(g.nams),1)
  prnt.lst <- list(p.vals=p.vals, g.nams=g.nams)
  # calculate means for file tables
  dims <- dim(av$data)
  loop.dim <- c(1:(model.info$MSav.ncol+1))
  dims <- dims[loop.dim]
  #print('calc vals')
  #sdvs <- apply(av$data, loop.dim, mysd)
  # if we selected moment stabilization this gets reverted here to write the "original"
  # averages to the output files.
  if(info$mom.stab)
    array.data <- data.rest.g.moments(av$data, info)
  else
    array.data <- av$data
  mns <- apply(array.data, loop.dim, mean)
  #print('sort terms')
  dim(mns) <- dims
  #dim(sdvs) <- dims
  f.ret <- order.4.file(p.vals, is.up, g.nams, mns, data.colnams, base.level,
                         contrast.name)
  #print(names(f.ret))
  #print(dim(data))
    # print(is.factor(f.ret$p.val))
  ret <- list(file.if=f.ret, prnt.if=prnt.lst, n.comps=n.comps)
  #print(names(ret$file.if))
  #print('bye')
  ret
}

is.cntr <- function(p.val)
  # cntr is integer!
  {
    (p.val==round(p.val)) & (p.val > 1)
  }

ou.int.file <- function(ou.ret, fname, p.type=p.adjust.methods, p.thrs, n.comps)
# write a table of significant genes to file fnam.
{
  # for p-value adjustment
  require(stats)
  # check whether we have a count
  do.cnt <- is.cntr(p.thrs)
  # adjust p.values
  ou.ret$p.val <- p.adjust(ou.ret$p.val, method=p.type, n=n.comps)
  min.P <- ou.ret$p.val
  if (!do.cnt)
    {
      se.dx <- min.P<=p.thrs
    }
  else
    {
      se.dx <- rep(F, length(min.P))
      p.thrs <- min(c(p.thrs, length(min.P)))
      se.dx[1:p.thrs] <- T
    }
  ou.ret <- ou.ret[se.dx,]
  write.table(ou.ret, file = fname, sep = "\t", row.names=F, quote=F)
  T
}

get.n.comps <- function(RG, info)
  # count number of comparisons in all contrasts
  {
    n.genes <- length(RG$gene)
    n.base.cols <- length(info$rank.names) - 1
    n.comp <- 0
    if (info$do.anova.ranking) # we have n.genes comparisons there
      n.comp <- n.comp+n.genes
    if (info$do.variety)
      n.comp <- n.comp+n.genes      
    # info$contrast is a list and we go through it to count all comparisons
    n.contr <- length(info$contrast)
    n <- 1
    while(n <= n.contr)
      {
        if (length(info$contrast[n])==1) # then we have multiple pairwise comparisons here
          n.comp <- n.comp + n.genes*n.base.cols
        else # classical contrast and only n.genes comparisons
          n.comp <- n.comp + n.genes
        n <- n+1
      }
    n.comp
  }

contrast.rank.2.file <- function(av, RG, model.info, info,
                                 p.type=p.adjust.methods,
                                 p.thrs, n.comps, log.fname = fspma.logfile)
  # writes a rank list using the previously implemented functions we provide
  # at the moment
  {
    fname <- info$ctrfname
    if (!any(fname=='NA'))
      {
        all.contrasts <- info$contrast
        all.nams <- names(all.contrasts)
        p.type <- info$p.type
        p.thrs <- info$p.thrs
        # now we loop over the contrasts
        n.contr <- length(all.contrasts)
        #print(n.contr)
        rank.res.4.plt <- list()
        n <- 1
        while(n<=n.contr)
          {
            contrast <- all.contrasts[[n]]
            # print(contrast)
            contrast.name <- all.nams[n]
            if(length(contrast)==1)  # we need contrast names only for baslevel tests otherwise get duplications
              cfnam <- paste(contrast.name, fname, sep='')
            else
              cfnam <- fname
                #print(contrast.name)
                # check whether we have a new type contrast or extract base level
            nc <- length(contrast)
            if (nc > 1)
              base.level <- NULL
            else
              {
                base.level <- contrast
                contrast <- NULL
              }
                # get names for mean table
            data.colnams <- info$rank.names
                # get equation describing rank term interaction from model.info
            gt.eff <- model.info$eq.eff
                # get relevant random effects interaction from model.info
            if (model.info$eq.rand != ~1) # if we do have a random effects model
              rand.eff <- model.info$eq.rand # use the relevant equation
            else
              rand.eff <- NULL  # otherwise we use NULL for fixed effects calculation
                # now we call the calculation
            if (is.null(contrast)) # then ou.int.get should use the base.level to deduce names.
              contrast.name <- NULL
            if(!is.null(base.level))
              strg <- sprintf('base: %3.0f - shortcut for several contrasts, be patient!\n', base.level)
            else
              strg <- sprintf('contrast: %s \n',
                              paste(contrast[1], paste("", contrast[-1], sep=",", collapse=""), sep="", collapse=""))
            stream.log(strg, fname=log.fname)
            ou.res <- ou.int.get(av, RG, model.info, info, data.colnams, gt.eff,
                                 rand.eff, base.level, contrast, contrast.name)
                # and finally loop over all comparisons and write result files
            allf.res <- ou.res$file.if
            all.cnams <- names(allf.res)
            all.fnams <- paste(all.cnams, cfnam, sep='.')
            if(is.null(contrast))  # we need to add the contrast names for baslevel otherwise get duplicates
              {
                all.cnams <- paste(all.cnams, all.nams[n], sep='.') # concatenate with baselevel name
                names(allf.res) <- all.cnams
              }
            # and now attach to all contrast list.
            rank.res.4.plt <- c(rank.res.4.plt, allf.res)
                #print(all.fnams)
            m.res <- length(allf.res)
            #print(m.res)
            m <- 1
            while(m<=m.res)
              {
                ou.int.file(allf.res[[m]], all.fnams[m], p.type, p.thrs, n.comps)
                m <- m+1
              }
            n <- n+1
          }
      }
    #print('bye2')
    rank.res.4.plt
  }

# helper that calculates significance levels of differential expression
# in varieties (i.e. in log ratio of channels). This is based on the
# original parametric version in YASMA (without bootstrap sampling).
fspma.pval.channel <- function(anv, info)
{
  a <- anv$data
  # YASMA!
  sd <- sd.g(anv)
  #print(sd)
  # yasma
  df <- df.g(anv$data)
  #print(df)
  av.log.diff <- diff.g(a)
  #print(av.log.diff)
  is.up <- av.log.diff > 0
  mn.diff <- abs(av.log.diff)/sd
  # p - value of differential expression.
  p.vals <- 2*pt(mn.diff, df=df, lower.tail=F)
  #print(p.vals)
  all.dims <- c(length(p.vals), 1)
  dim(p.vals) <- all.dims
  dim(is.up) <- all.dims
  dim(av.log.diff) <- all.dims
  # if we selected moment stabilization this gets reverted here to write the "original"
  # averages to the output files.
  if(info$mom.stab)
    array.data <- data.rest.g.moments(anv$data, info)
  else
    array.data <- anv$data
  mns <- diff.g(array.data)
  dim(mns) <- all.dims
  list(p.vals=p.vals, is.up=is.up, mns=mns)
}

do.rank.variety <- function(av, RG, model.info, info, n.comps)
  # rank all genes w.r.t. average log ratio between Cy5 and Cy3.
{
  ret <- fspma.pval.channel(av, info)
  #print(names(ret))
  f.res <- order.4.file(ret$p.vals, ret$is.up, RG$gene,ret$mns, 'Cy5.Cy3',
                        contrast.name='variety')
  cnam <- names(f.res)
  fnam <- paste(cnam, info$ctrfnam, sep='.')
  # fnam <- paste(cnam, fnam, sep='.')
  #print(f.res)
  ou.int.file(f.res[[1]], fnam, info$p.type, info$p.thrs, n.comps)
  f.res
}

stack.reps <- function(RG, info)
  {
    n.rep <- info$effects[length(info$effects)]
    newdim <- dim(RG$R)
    newdim[1] <- n.rep*newdim[1]
    newdim[2] <- newdim[2]/n.rep
#    print(newdim)
#    print(dim(RG$R))
    dim(RG$R) <- newdim
    dim(RG$Rb) <- newdim
    dim(RG$G) <- newdim
    dim(RG$Gb) <- newdim
    dim(RG$Flg) <- newdim
    dim(RG$x) <- newdim
    dim(RG$y) <- newdim
    dim(RG$gridno) <- newdim
    dim(RG$slideno) <- newdim
    RG
  }

unstack.reps <- function(RG, info)
  {
    n.rep <- info$effects[length(info$effects)]
    newdim <- dim(RG$R)
    newdim[1] <- newdim[1]/n.rep
    newdim[2] <- newdim[2]*n.rep
#    print(newdim)
#    print(names(RG))
    dim(RG$R) <- newdim
    dim(RG$Rb) <- newdim
    dim(RG$G) <- newdim
    dim(RG$Gb) <- newdim
    dim(RG$Flg) <- newdim
    dim(RG$x) <- newdim
    dim(RG$y) <- newdim
    dim(RG$gridno) <- newdim
    dim(RG$slideno) <- newdim
    RG
  }

# The follwoing code allows to normalize data within slide (i.e. the RG structure is reshaped
# to within slide replicates being treated as different genes) This is necessary since
# YASMAs normalization functions assume that only genes are in a slide. Replicates are not available.).
reshapeRG <- function(rg, experiments, genes)
  {
# reshaeRG allws to sdjust the RG structure w.r.t. a new experiments setting
# They only assumption the function makes is that prod(dim(RG$R)) etc. matches
# the number of experiments/genes required. The function ought to be used
# with care since it is possible to mess up the data and thus get a completely
# wrong analysis. It is used to move between techical replicates an slides
# which we need to perform a sensible between slide normalization.
    # set genes
    n.genes <- length(genes)    
    # number of experiments in original datastructure
    n.exp.old <- ncol(rg$R)
    n.vars <- length(experiments) # number of effects (without genes!)
    # in new experiment
    # and from experiments we get the new number of levels in all
    # effects (again without genes) using unlist (drop the $)
    # The only way of doing it general without knowing the names of the
    # effects.
    n.exp <- 1
    for( i in 1:n.vars)
      n.exp <- n.exp * length(unlist(experiments[i], use.names=F))
    # finally we check whether we can redim the data.
    # if this goes wrong the new specification was faulty.
    n.old <- prod(dim(rg$R))
    n.new <- n.exp*n.genes
    if (n.old != n.new)
      stop("New experiments/genes missmatch data in RG structure")
    # beyond that we can safely modify the RG.
    rg$genes <- genes
    dim(rg$R) <- c(n.genes, n.exp)
    dim(rg$Rb) <- c(n.genes, n.exp)
    dim(rg$G) <- c(n.genes, n.exp)
    dim(rg$Gb) <- c(n.genes, n.exp)
    rg$experiments <- experiments
    rg
  }

parse.norm.type <- function(info)
  {
    # parse norm.type to obtain
    # normalization methods including parameters and
    # to get scale flag (make scale last operation irrespective
    # of where it is found in the list)
    norm.ret <- list()
    do.scale <- F
    n.norm <- length(info$norm.type)
    n <- 1
    have.loess <- F
    while (n<=n.norm)
      {
        if (have.loess) 
          {
            if (is.finite(p.val <- as.numeric(info$norm.type[n])))
              {
                this.norm[p.cntr] <- p.val
                p.cntr <- p.cntr+1
              }
            else
              {
                if (this.norm$deg>2 | this.norm$deg <1)
                  this.norm$deg <- 1
                if (this.norm$span>1 | this.norm$span < 0)
                  this.norm$span <- 0.9
                norm.ret <- c(norm.ret, list(this.norm))
                have.loess <- F
              }
          }
        if (info$norm.type[n]=='loess')
          {
            this.norm <- list(type='loess', span=0.9, deg=1)
            have.loess <- T
            p.cntr <- 2
          }
        if (info$norm.type[n]=='location')
          {
            this.norm <- list(type='location')
            norm.ret <- c(norm.ret, list(this.norm))
          }
        if (info$norm.type[n]=='scale')     
          do.scale <- T
        n <- n+1
      }
    if (have.loess)
      {
        if (this.norm$deg>2 | this.norm$deg <1)
          this.norm$deg <- 1
        if (this.norm$span>1 | this.norm$span < 0)
          this.norm$span <- 0.9
        norm.ret <- c(norm.ret, list(this.norm))
        have.loess <- F
      }
    list(do.scale=do.scale, all.norm=norm.ret)
  }

do.normalize <- function(RG, info, log.fname=fspma.logfile)
  {
    # reshape RG structure such that within slide replicates are mapped
    # to additional genes. (to get slide based normalization instead of
    # gene set based normalization).
    # performs normalization as controlled by info$norm.type
    # IMPORTANT: the order is important if loess is used together with scale
    # FIRST loess and THEN scale. (MUST be correct in def file !!)
    if (!any(info$norm.type == 'NA') & !any(info$norm.type=='spike'))
      # we use NA to indicate "not applicable", spikes are handeled elsewhere.
      {
        # first reshape RG:
        orig.exp <- RG$experiments
        orig.gene <- RG$genes
        n.eff <- length(orig.exp)
        on.slide.rep <- info$effects[n.eff] # on slide replication
        if (on.slide.rep > 1) # then we have to reshape to get slide based normalization.
          {
            RG <- stack.reps(RG, info)
            #genes <- rep(orig.gene, on.slide.rep)
            #n.exp <- orig.exp[1:(n.eff-1)]
            #RG <- reshapeRG(RG, n.exp, genes)
          }
        # now we are ready to do the normalization bit
        RG.n <- RG
        # parse norm.type
        norm.type <- parse.norm.type(info)
        n.norm <- length(norm.type$all.norm)
        n <- 1
        while (n <= n.norm)
          {
            stream.log(sprintf("Normalising by: %s\n", norm.type$all.norm[[n]]$type), fname=log.fname)
            # note that we ovverride yasma's defaluts.
            # on one hand they do not allow log scale data in R and G.
            # further more yasma's implementation of loess does not forward
            # any degree and scaleparameter that is provided to the fucntion.
            RG.n <- switch(norm.type$all.norm[[n]]$type,
                           loess=fspma.logRG.loess.norm(RG.n, deg=norm.type$all.norm[[n]]$deg,
                             span=norm.type$all.norm[[n]]$span, show=F, tolog=info$dologarray),
                           location=fspma.logRG.shift.norm(RG.n, tolog=info$dologarray))             
            n <- n+1
          }
        if (norm.type$do.scale)
          {
            stream.log("Normalising by: scale\n", fname=log.fname)
            RG.n=fspma.logRG.scale.norm(RG.n, tolog=info$dologarray)
          }
        RG$R <- RG.n$R
        RG$Rb <- RG.n$Rb
        RG$G <- RG.n$G
        RG$Gb <- RG.n$Gb
        if (on.slide.rep > 1) # then we have to reshape to get back to original layout
          {
            RG <- unstack.reps(RG, info)
            #RG <- reshapeRG(RG, orig.exp, orig.gene)
          }
      }
    RG
  }

# The next functions use ANOVA models to get p-values for entire
# time series (or other designs).
myappl.bfaov <- function(data, eqn, new.names, MS.ncol, MS.dcol)
  # calculates one p-value entry in an anova table for gene wise
  # ranking of longitudinal microarray experiments with 
  # random effects. allows to use bfaov in connection with
  # apply on the YASMA compatible data structure.
   {
     # print(dim(data))
     # 06 2008: the following patches R >=2.62 which strips dim names when using apply
     if (length(names(dim(data)))==0)
         names(dim(data)) <- new.names
     end
     # first calculate anova
     bf.obj <- bfaov(eqn, data)
     # then extract the p-value that corresponds to the
     # grouping we want to test for. The indices are obtained from
     # that piece of code that generates now all model formulae etc.
     # from infromation that is easy to provide.
     tmp <- anova.from.comblist(bf.obj)
     #print(tmp)
     # MS.dcol is 0 in case we have no random effects.
     if (MS.dcol !=0) # then we have a random effect
       {
         MSd <- tmp$MS[MS.dcol]
         dfd <- tmp$df[MS.dcol]
       }
     else # no random effect and we get MSE from calling anova
       {
         ret <- anova(bf.obj, ptype='none')
         MSd <- ret$MS
         dfd <- ret$df
       }
     MSn <- tmp$MS[MS.ncol]
     dfn <- tmp$df[MS.ncol]
     #print(list(MSd, MSn, dfd, dfn))
     F.stat <- MSn/MSd
     p <- 1 - pf(F.stat,dfn,dfd)
     list(p=p, F.stat=F.stat)
   }
#data <- l.tr.logrt[1,,,]
#dims <- c(7, 4, 2)
#names(dims)<- c('time', 'smpl', 'trep')
#dim(data) <- dims
#data
#bf <- myappl.bfaov(data, ret$eq.avrnk, ret$MSav.ncol, ret$MSav.dcol)

calc.gene.anova <- function(log.rt.data, model.info)
  # calculate ANOVA tables for anova.eqn for the data in log.rt.data
  # the function calculates for each data subset of log.rt.data
  # an ANOVA table, which is controlled using model.info
  {
    if (model.info$MSav.ncol<=1)
      {
        loop.dim <- 1
        max.dim <- 1
      }
    else
      {
        loop.dim <- c(1:(model.info$MSav.ncol-1))
        max.dim <- model.info$MSav.ncol-1
      }
    all.names <- names(dim(log.rt.data))
    len.names <- length(all.names)
    new.names <- all.names[(max.dim+1):len.names]
    #print(new.names)
    # print(loop.dim)
    # now we loop over the loop factor, cut out the corresponding subset of
    # log.rt.data, calculate the anova table and extract F-statistics and p-values
    # for the specification previously obtained from MS between specified by
    # model.info, obtained from calling gen.yasma.eqns.
    #print(names(dim(log.rt.data)))
    #print(model.info$eq.avrnk)
    #print(loop.dim)
    #print(model.info$MSav.ncol)
    bfaov.list <- apply(log.rt.data, loop.dim, myappl.bfaov, model.info$eq.avrnk, new.names,
                        model.info$MSav.ncol, model.info$MSav.dcol)
  }

#av.rnk.vals <- calc.gene.anova(l.tr.logrt, model.info)

filter.gene.anova <- function(bfaov.list, genenams,
                              p.type=p.adjust.methods, p.thrs, n.comps)
  # filter anova ranking
  {
    allP <- NULL
    allF <- NULL
    n.val <- length(bfaov.list)
    # collect values from result list
    for(n in 1:n.val)
      {
        allP <- rbind(allP, bfaov.list[[n]]$p)
        allF <- rbind(allF, bfaov.list[[n]]$F.stat)
      }
    allP <- p.adjust(allP, method=p.type, n=n.comps)
    # sort w.r.t. p-values (increasing)
    ret <- sort(allP, index=T)
    allP <- ret$x
    dx <- ret$ix
    genenams <- genenams[dx]
    allF <- allF[dx]
    if (!is.cntr(p.thrs))
      sel.bool <- allP<=p.thrs
    else # we hav a count
      {
        sel.bool <- rep(F, n.val)
        sel.bool[1:p.thrs] <- T
      }
    allP <- allP[sel.bool]
    genenams <- genenams[sel.bool]
    dx <- dx[sel.bool]
    allF <- allF[sel.bool]
    ret <- data.frame(dx, I(genenams), allP) #, allF)
    names(ret) <- c('dx', 'g.nams', 'p.vals')
    ret
  }

#tc.av.rnk <- filter.gene.anova(av.rnk.vals, new.RG$genes, 'bonf', 0.05)
    
write.tcdat.diff.expr <- function(data, av.rnk.filter, model.info,
                                  collapse.names, fname='av.express.tsv')
  # collapse time series data and write into file with rank info and gene names
  {
    dims <- dim(data)
    loop.dim <- c(1:(model.info$MSav.ncol+1))
    dims <- dims[loop.dim]
#    print(dims)
#    sdev <- apply(data, loop.dim, mysd)
    data <- apply(data, loop.dim, mean)
#    print(length(sdev))
#    print(length(data))
    dim(data) <- dims
   # dim(sdev) <- dims
    #print(dim(data))
    if (length(loop.dim) > 2)
      {
        nlen <- length(av.rnk.filter$dx)
        ndat <- NULL
        for (n in 1:nlen)
          {
            #nsdev <- rbind(nsdev, sdev[slice.index(data, 1)==av.rnk.filter$dx[n]])
            ndat <- rbind(ndat, data[slice.index(data, 1)==av.rnk.filter$dx[n]])
          }
        data <- ndat
        #sdev <- sdev
      }
    else
      {
        data <- data[av.rnk.filter$dx, ]
        #sdev <- sdev[av.rnk.filter$dx, ]
      }
    #print(dim(data))
    #print(length(data))
    names.frame <- names(av.rnk.filter)
    #print(names.frame)
    data.4.file <- cbind(av.rnk.filter, data.frame(data))
    names.frame <- c(names.frame, collapse.names)
    #print(dim(data.4.file))
    names(data.4.file) <- names.frame
    write.table(data.4.file, file = fname, sep = "\t", row.names=F, quote=F)
    data.4.file
  }

# collapse.names <- c('mn adlt','t1','t2','t3','t4','t5','t6', 'sd adlt','t1','t2','t3','t4','t5','t6')
# write.tcdat.diff.expr(l.tr.logrt, tc.av.rnk, model.info, collapse.names, 'rfav.mn.bonf.dat')

fspma.aovrnk <- function(info, model.info, array.data, fname=NULL, n.comps)
  # interface between fsma info and model.info structure and the
  # functions for ANOVA based ranking.
  {
    tab.ret <- NA
    if (is.null(fname))
      fname <- info$ctrfname
    if (!any(fname=='NA')) # then we calculate and write the ANOVA based ranking
      {
        fname <- paste('anova.rank.', fname, sep='')
        # first calculate the bfaov.list (i.e. list of ANOVA models)
        bfaov.list <- calc.gene.anova(array.data, model.info)
        # then rank the list
        rank.list <- filter.gene.anova(bfaov.list, info$geneids, info$p.type, info$p.thrs, n.comps)
        # get the names for the mn and std. dev columns
        mn.nams <- paste(info$rank.names, 'mean', sep='.')
        #sd.nams <- paste(info$rank.names, 'std.dev',sep='.')
        collapse.names <- c(mn.nams)#, sd.nams)
        # if we selected moment stabilization this gets reverted here to write the "original"
        # averages to the output files.
        if(info$mom.stab)
          array.data <- data.rest.g.moments(array.data, info)
        # and write to file
        tab.ret <- write.tcdat.diff.expr(array.data, rank.list, model.info,
                                         collapse.names, fname)
      }
    list(anova.ranklist=tab.ret) # return the data frame that is written to the file
  }

knn.dis <- function(x, val, cols)
  # calculate squared euclidian distance and avoid smaller eps
  # (otherwise problems with imputing!)
  {
    max(fspma.eps, sum((x[cols]-val[cols])^2))
  }

knn.array.impute <- function(array.data, k, max.iter=50)
  # impute array data with knn
  {
    # generate current solution
    na.dx <- is.na(array.data)
    dim(na.dx) <- dim(array.data)
    array.data[na.dx] <- 0
    isnan.rws <- array(apply(na.dx, 1, any), c(nrow(array.data),1))
    # print(dim(isnan.rws))
    n.rws <- length(isnan.rws)
    all.rws <- c(1:n.rws)
    nan.rws <- all.rws[isnan.rws]
#    print(nan.rws)
    do.calc <- T
    iter <- 1
    rem.rws <- NULL
    cat('\n')
    while (do.calc)
      {
        old.arr <- array.data
        for (n in nan.rws)
          {
            cat('.')
            # this version calculates from all rows
            calc.rws <- all.rws[all.rws!=n]
            # first get all Euclidian distances from those data without nan's
            # to this nan-entry (ignoring the nan's)
            col.impute <- na.dx[n,]
            col.calc <- !col.impute
            if (any(col.calc)) # then impute
              {
                dist <- apply(array.data[calc.rws,], 1, knn.dis, array.data[n,],
                          col.calc)
                # print(length(dist))
                # print(sum(!is.finite(dist)))
                # now sort the distances
                rt <- sort(dist, index.return=T)
                ds <- array(rt$x, c(length(dist),1))
                # invert them since close profiles should get more weight
                ds <- ds^(-1)
                # impute.rws are the k - nearest neighbours to the profile that is imputed
                impute.rws <- calc.rws[rt$ix[1:k]]
                impute.wght <- ds[1:k, rep(1, sum(col.impute))]/sum(ds[1:k])
                # print(sum(col.impute))
                dim(impute.wght) <- c(k, sum(col.impute))
                # print(dim(impute.wght))
                # print(dim(ds))
                # print(colSums(impute.wght))
                # print(impute.wght)
                # and impute
                val.impute <- colSums(array.data[impute.rws, col.impute]*impute.wght)
                #print(val.impute)
                array.data[n,col.impute] <- val.impute
              }
            else # add this row (gene) to rem.rws
              {
                rem.rws <- c(rem.rws, n)
              }
          }
        cur.dist <- sum((old.arr-array.data)^2)
        do.calc <- cur.dist>fspma.knn.max.tol & iter < max.iter
        iter <- iter+1;
        # and remove rem.rws from nan.rws (since they can't be imputed)
        if (!is.null(rem.rws))
          {
            nan.rws <- nan.rws[nan.rws!=rem.rws]
            rem.rws <- NULL
          }
        # remark: old.arr gets assigned at the top of while loop!
        cat('\n')
      }
    list(array.data=array.data, rem.rws=rem.rws)
  }
        
# functionality for treating missing or invalid spots flags
knn.impute <- function(RG, k=10, go2log=T)
  # imputes missing values by KNN
  {
    # we first get missing values 
    na.dx <- is.na(RG$R)|is.na(RG$G)|is.na(RG$Rb)|is.na(RG$Gb)
    if (sum(na.dx)>0) # there is something to impute
      {
        # na.rws <- array(apply(na.dx, 1, any), c(nrow(RG$R),1))
        # convert R channel data to log expression values
        if (go2log)
          {
            R.dt <- RG$R-RG$Rb
            R.dt[R.dt<fspma.eps] <- fspma.eps
            lg.R <- log2(R.dt)
          }
        else
          lg.R <- RG$R
        lg.R[na.dx] <- NA
        # impute on lg.R
        ret <- knn.array.impute(lg.R,  k)
        lg.R <- ret$array.data
        rem.rws <- ret$rem.rws
        #print(sum(is.na(lg.R)))
        # convert G channel data to log expression values
        if (go2log)
          {
            G.dt <- RG$G-RG$Gb
            G.dt[G.dt<fspma.eps] <- fspma.eps
            lg.G <- log2(G.dt)
          }
        else
          lg.G <- RG$G
        lg.G[na.dx] <- NA
        # impute on lg.G
        ret <- knn.array.impute(lg.G, k)
        lg.G <- ret$array.data
        rem.rws <- unique(c(rem.rws, ret$rem.rws))
        #print(sum(is.na(lg.G)))
        # and now restore RG (this implies removing both backgrounds)
        dim(lg.R) <- dim(RG$R)
        if (go2log)
          {
            RG$R <- 2^lg.R
            RG$Rb <- array(rep(0, prod(dim(RG$R))), dim(RG$R))
          }
        else
          RG$R <- lg.R
        dim(lg.G) <- dim(RG$G)
        if(go2log)
          {
            RG$G <- 2^lg.G
            RG$Gb <- array(rep(0, prod(dim(RG$G))), dim(RG$G))
          }
        else
            RG$G <- lg.G
        if (!is.null(rem.rws)) # then some of the genes are all NA and get removed from RG
          {
            all.rws <- rep(T, nrow(RG$G))
            all.rws[rem.rws] <- F
            RG$R <- RG$R[all.rws,]
            RG$G <- RG$G[all.rws,]
            RG$Rb <- RG$Rb[all.rws,]
            RG$Gb <- RG$Gb[all.rws,]
            RG$Flg <- RG$Flg[all.rws,]
            if(any(names(RG)=='x')) # then we have a version 2 thing
              {
                RG$x <- RG$x[all.rws,]
                RG$y <- RG$y[all.rws,]
                RG$slideno <- RG$slideno[all.rws,]
                RG$gridno <- RG$gridno[all.rws,]
              }
            RG$genes <- RG$genes[all.rws]
          }
      }
    #print(dim(RG$R))
    #print(dim(RG$Rb))
    #print(dim(RG$G))
    #print(dim(RG$Gb))
    #print(dim(RG$Flg))
    #print(dim(RG$x))
    #print(dim(RG$y))
    #print(dim(RG$slideno))
    #print(dim(RG$gridno))    
    RG
  }

del.impute <- function(RG)
  {
    # we first get missing values 
    na.dx <- is.na(RG$R)|is.na(RG$G)|is.na(RG$Rb)|is.na(RG$Gb)
    if (sum(na.dx)>0) # there is something to impute
      {
        na.rws <- array(apply(na.dx, 1, any), c(nrow(RG$R),1))
        RG$R <- RG$R[!na.rws,]
        RG$G <- RG$G[!na.rws,]
        RG$Rb <- RG$Rb[!na.rws,]
        RG$Gb <- RG$Gb[!na.rws,]
        RG$Flg <- RG$Flg[!na.rws,]
        RG$genes <- RG$genes[!na.rws]
        if(any(names(RG)=='x')) # then we have a version 2 thing
          {
            RG$x <- RG$x[!na.rws,]
            RG$y <- RG$y[!na.rws,]
            RG$slideno <- RG$slideno[!na.rws,]
            RG$gridno <- RG$gridno[!na.rws,]
          }
      }
    #print(dim(RG$R))
    #print(dim(RG$Rb))
    #print(dim(RG$G))
    #print(dim(RG$Gb))
    #print(dim(RG$Flg))
    #print(dim(RG$x))
    #print(dim(RG$y))
    #print(dim(RG$slideno))
    #print(dim(RG$gridno))    
    RG
  }

treat.na <- function(RG, method, go2log=T)
  {
    # do treat NA by removing genes, imputing by knn or not treat them at all
    # method is equivalent to info$impute.mthd which is a list. Pos 1 is either
    # 'knn' or 'del' and position 2 is k, the number of nearest neighburs to be
    # considered by knn.impute.
    switch(method[[1]],
           del=del.impute(RG),
           knn=knn.impute(RG, method[[2]], go2log),
           RG)
  }

cnt.na <- function(RG)
  {
    # counts NA's in RG called to check after imputing values.
    sum(is.na(RG$R)|is.na(RG$G)|is.na(RG$Rb)|is.na(RG$Gb)|is.na(RG$x)|is.na(RG$y))
  }

wrtna.2.file <- function(fname='na_debug.tsv', RG, info)
  {
    # writes NA debuging info to file.
    # This file should be considered for debugging NA information
    # first we generate a matrix:
    diag.rws <- length(RG$gene)
    n.rep <- info$effects[length(info$effects)]
    # print(rep)
    diag.cols <- ncol(RG$R)%/%n.rep # must work anyway!
    #diag.rws <- nrow(RG$R)
    na.diag <- array(rep(':', diag.rws*diag.cols), c(diag.rws, diag.cols))
    na.dx <- is.na(RG$R)|is.na(RG$G)|is.na(RG$Rb)|is.na(RG$Gb)|is.na(RG$x)|is.na(RG$y)|is.na(RG$gridno)
    if (n.rep > 1)
      {
        na.fdx <- array(rep(F, diag.rws*diag.cols), c(diag.rws, diag.cols))
        cc <- 0
        while (cc<diag.cols)
          {
            dx.cdx <- n.rep*cc + c(1:n.rep)
            na.fdx[,cc+1] <- apply(na.dx[, dx.cdx],1, any)
            cc <- cc+1
          }
      }
    else
      na.fdx <- na.dx
    na.diag[na.fdx] <- 'NA'
    # generate column names for the dataframe
    allfnams <- rep('', diag.cols)
    n <- 1
    while(n <= diag.cols)
      {
        dx <- comb2dx(info$dat.file.rd[[n]]$expdx, info$effects)
        allfnams[dx] <- paste('File pos:', n)
        n <- n + 1
      }
    na.data.frame <- data.frame(na.diag, row.names = info$geneids)
    names(na.data.frame) <- allfnams
    write.table(na.data.frame, file = fname, sep = "\t", row.names=T, quote=F)
  }

file.out.RG <- function(info, RG, wrtchn=F)
  # Writes data from RG structre into file. We write genes as column headers and
  # then all the RG entries. A second file contains the effects that we need for
  # various modeling techniques as well.
  {
    if (!wrtchn) # defult writes log ratios
      {
        # data (log ratios is written)
        dat.fnam <- paste(info$datoutfnam, 'logrt.tsv', sep='')
        if(info$dologarray)
          {
            R <- RG$R-RG$Rb
            idbeps <- R<fspma.eps
            R[idbeps] <- fspma.eps        
            G <- RG$G-RG$Gb
            idbeps <- G<fspma.eps
            G[idbeps] <- fspma.eps
            vals <- as.matrix(log2(R)-log2(G))
            vals <- as.data.frame(t(vals))
          }
        else
          {
            vals <- as.matrix((RG$R-RG$Rb)-(RG$G-RG$Gb))
            vals <- as.data.frame(t(vals))
          }
        names(vals) <- RG$genes
        write.table(vals, file = dat.fnam, sep = "\t", row.names=F, quote=F)
      }
    else # this can be overridden by wrtchn==TRUE, where we write log channel values.
      {
        # data (log channel values are written)
        if(info$dologarray)
          {
            R <- RG$R-RG$Rb
            idbeps <- R<fspma.eps
            R[idbeps] <- fspma.eps        
            G <- RG$G-RG$Gb
            idbeps <- G<fspma.eps
            G[idbeps] <- fspma.eps
            R.vals <- as.matrix(log2(R))
            G.vals <- as.matrix(log2(G))
            R.vals <- as.data.frame(t(R.vals))
            G.vals <- as.data.frame(t(G.vals))
          }
        else
          {
            R.vals <- as.matrix((RG$R-RG$Rb))
            G.vals <- as.matrix((RG$G-RG$Gb))
            R.vals <- as.data.frame(t(R.vals))
            G.vals <- as.data.frame(t(G.vals))
          }
        datR.fnam <- paste(info$datoutfnam, 'logR.tsv', sep='')
        datG.fnam <- paste(info$datoutfnam, 'logG.tsv', sep='')
        names(R.vals) <- RG$genes
        names(G.vals) <- RG$genes
        write.table(R.vals, file = datR.fnam, sep = "\t", row.names=F, quote=F)
        write.table(G.vals, file = datG.fnam, sep = "\t", row.names=F, quote=F)
      }
    # corresponding classification information (i.e replicate no, slide no etc.)
    eff.data <- as.data.frame(RG$eff.store)
    eff.nams <- paste(info$effnams, info$randeff, sep='_is_random_')
    names(eff.data) <- eff.nams
    eff.fnam <- paste(info$datoutfnam, 'effdesc.tsv', sep='')
    write.table(eff.data, file = eff.fnam, sep = "\t", row.names=F, quote=F)
  }

rem.na.xygrid <- function(RG)
  {
    # we first get missing values 
    na.dx <- is.na(RG$x)|is.na(RG$y)|is.na(RG$gridno)
    if (sum(na.dx)>0) # there is something to impute
      {
        na.rws <- array(apply(na.dx, 1, any), c(nrow(RG$R),1))
        cat('NA found - see debug info removing genes: ', RG$genes[na.rws], '\n')
        RG$R <- RG$R[!na.rws,]
        RG$G <- RG$G[!na.rws,]
        RG$Rb <- RG$Rb[!na.rws,]
        RG$Gb <- RG$Gb[!na.rws,]
        RG$Flg <- RG$Flg[!na.rws,]
        RG$genes <- RG$genes[!na.rws]
        if(any(names(RG)=='x')) # then we have a version 2 thing
          {
            RG$x <- RG$x[!na.rws,]
            RG$y <- RG$y[!na.rws,]
            RG$slideno <- RG$slideno[!na.rws,]
            RG$gridno <- RG$gridno[!na.rws,]
          }
      }
    RG
  }

# some functions that try to make the analysis numerically less problematic.
# We suggest a gene wise adjustment of the first two moments whcih is stored in.
# info and removed when writing the data.
data.rem.g.moments <- function(av.data, info)
  {
    # input: av.data - which is the log ratio conversion from yasma
    #        info - fspmas info structure, which is modified as well to store
    #               the normalization information.
    # returns: list(av.data, info)
    # the normalized log expression or log ratio values and info, which contains the
    # removed moments.
    ddx <- c(1:length(dim(av.data)))
    along.dim <- ddx[names(dim(av.data))=='G']
    g.mn <- apply(av.data, along.dim, mean)
    # removing the mean uses R's way of reusing arguments.
    # i.e. the next line implicitly replicates the estimated moment.
    av.data <- av.data-g.mn
    # same for standard dev.
    g.sd <- apply(av.data, along.dim, mysd)
    av.data <- av.data/g.sd
    tmpl <- list(g.mn=g.mn, g.sd=g.sd)
    info <- c(info, tmpl)
    list(av.data=av.data, info=info)
  }

data.rest.g.moments <- function(av.data, info)
  {
    # input: av.data - which is the log ratio conversion from yasma
    #        info - fspmas info structure, which contains the previously 
    #               calculated normalization information.
    # returns: av.data
    # the restored log expression or log ratio values
    
    # restoring the mean uses R's way of reusing arguments.
    # i.e. the next line implicitly replicates the estimated moment.
    # same for standard dev.
    av.data <- av.data*info$g.sd
    av.data <- av.data+info$g.mn
    av.data
  }

check.geneids <- function(info, RG)
  {
    ig <- info$geneids
    rg <- RG$genes
    li <- length(ig)
    lr <- length(rg)
    g.ok <- li==lr
    n <- 1
    while (g.ok & n <=li)
      {
        g.ok <- ig[n]==rg[n]
        n <- n + 1
      }
    g.ok
  }

fspma.wrapper <- function(def.fname=NULL, RG=NULL, info=NULL, na.debug.fname='na_debug.tsv',
                          log.fname=fspma.logfile, init.logfile=T)
  # fspma.wrapper is a simple interface to and generalisation of the YASMA package that allows
  # ANOVA and test based inference for rather general microarray experiments.
  # The function is controlled via a definition file that is adaptable to
  # every balanced experiment design and slide data. It covers two color arrays
  # with and without background signals and one colour data like affymetrix and
  # meta data that were normalized and preprocessed elsewhere.
  # Details of such a definition file are provided with the library for
  # time course data with technical replicates as well as on slide replication.
  # Extensions: gene wise ANOVA ranking, two methods for delaing with bad spot flags:
  # remove gene and impute by knn. (The anova based approach provided with YASMA does
  # not work due to the enormous memory requirement of the linear model.)
  # IN:
  # def.fname - (definition file name)
  # RG        - if RG is NOT NULL the function does not load the data according to the def file,
  #             however all other modifications are taken from the def file.
  #             This allows for an efficient change in comparisons, normalizations etc.
  #             It is however at the users responsibility that the data and the description in
  #             the def file match. All modifications that influence data loading Potentially
  #             even apparently unrelated modifications like changing the
  #             specification of random effects and definitely modifications in the colour swap
  #             or similar make a new loading step strictly necessary. BE CAREFUL!!!!!
  # info      - info structure otherwise read from definition file. This allows
  #             together with the Load.Only: entry in the definition file to include other
  #             processing steps on a scripting level and to repeat processing (e.g. with different tests) 
  #             without need for several definition files.
  # Modifying BOTH PARAMETERS (IN PARTICULAR info) REQUIRE TO UNDERSTAND WHAT THE LIBRARY DOES AND ARE
  # SUGGESTED TO BE USED BY EXPERTS ONLY! 
  # OUT:
  # if info$load.only == TRUE:
  #
  # ret$info        - info from def file
  # ret$RG          - data loadid according to def file
  #
  # after processing:
  # 
  # ret$info        - info from def file
  # ret$RG          - data loadid according to def file
  # ret$model.info  - model description like equations etc.
  # ret$aov.a       - array data (from RG)
  # ret$av          - bfaov object
  #
  # (C) P. Sykacek 2004
  {
    #cat('new version !!!!!!!')
    if (init.logfile)
      {
        fid <- file(log.fname, "w")
        close(fid)
      }
    strg <- paste(paste("Run started on - ", format(Sys.time(), "%a %b %d %Y %X"), sep="", collapse=""),
                  "\n\n", sep="", collapse="")
    stream.log(strg, fname=log.fname)
    if (!is.null(def.fname))
      {
        stream.log('Loading info from file.\n', fname=log.fname)
        # here no initialisation since this is done by fspma.wrapper.
        info <- read.defs(def.fname, log.fname, F, F)
      }
    if(is.null(def.fname)&&is.null(info))
      {
        stream.log('Either definition file or info is needed for processing! Execution aborted.\n',
                   fname=log.fname)
        stop("Execution aborts!")
      }
    # then use the info structure to read all data (takes a while!)
    if(is.null(RG))
      {
        stream.log('Loading RG data.\n', fname=log.fname)
        if(info$fast.load)
          stream.log('Warning: using option fast load.\n', fname=log.fname)
        RG <- tabdel2rg(info)
      }
    else
      {
        stream.log('RG provided data not loaded.\n', fname=log.fname)
        # here we need to check whether RG genes are the same as info geneids.
        # if RG is loaded from files this is taken care of but here we may experience
        # differnces
        if (!check.geneids(info, RG))
          {
            stream.log('RG$genes and info$geneids mismatch. RG overrides info!\n', fname=log.fname)
            info$geneids <- RG$genes
          }
      }
    # check for any NA entries
    no.na <- cnt.na(RG)
    if (no.na>0)
      {
        strg <- sprintf("Number of NA entries from channel: %3.0f, see %s for more information.\n",
                        no.na, na.debug.fname)
        stream.log(strg, fname=log.fname)
        # if there are then write the info to the debug file name.
        wrtna.2.file(na.debug.fname, RG, info)
      }
    else
      {
        strg <- sprintf("Number of NA entries from channel: %3.0f.\n",no.na)
        stream.log(strg, fname=log.fname)
      }
    # now check for NA in x, y and gridno and if there are - eliminate the
    # genes from the analysis.
    RG <- rem.na.xygrid(RG)
    if(length(info$geneids)!=length(RG$genes))
      info$geneids <- RG$genes
    # finally normalize the positions
    RG <- norm.space.lin(RG)
    if (info$load.only)
      {
        info$load.only <- F
        ret <- list(info=info, RG=RG)
      }
    else
      {
        # first we treat missing or failed spot flags and imputing:
        if(!any(is.na(info$flag.nack)))
          {
            # first set flags for all nack values (there might be more!)
            set.na.dx <- RG$Flg==info$flag.nack[1]
            set.na.spk.dx <- RG$normspike$Flg == info$flag.nack[1]
            n.flagvals <- length(info$flag.nack)
            n <- 2
            while(n<=n.flagvals)
              {
                set.na.dx <- set.na.dx | RG$Flg==info$flag.nack[n]
                set.na.spk.dx <- set.na.spk.dx & RG$normspike$Flg == info$flag.nack[n]
                n <- n+1;
              }
            RG$R[set.na.dx] <- NA
            RG$G[set.na.dx] <- NA
            RG$normspike$Flg <- rep(1, length(RG$normspike$Flg))
            RG$normspike$Flg[set.na.spk.dx] <- NA
          }
        else
          set.na.spk.dx <- 0
        # number of any NA
        strg <- sprintf("Number of NA entries from flag: %3.0f (NA spikes): %3.0f\n", cnt.na(RG), sum(set.na.spk.dx))
        stream.log(strg, fname=log.fname)
        if(cnt.na(RG)>0 & info$impute.mthd=='NA')
          {
            stream.log("Warning: No imputation specified and NA's found!\n Override NA with del, otherwise we crash.\n", fname=log.fname)
            info$impute.mthd <- 'del'
          }
        # now we treat NA according to info$impute.mthd
        RG <- treat.na(RG, info$impute.mthd, info$dologarray)
        # and reset the gene information in info$geneids (some might have gone!)
        if(length(info$geneids)!=length(RG$genes))
          info$geneids <- RG$genes
        # after imputing we check how many NA's are there
        strg <- sprintf("Number of NA entries after imputing: %3.0f\n", cnt.na(RG))
        stream.log(strg, fname=log.fname)
        # and generate model information (equations etc.) from the info structure.
        stream.log('Getting model.info\n', fname=log.fname)
        model.info <- info.2.yasmaeqns(info)
        if(info$version >= 2) # then add the spike model equations
          model.info <- get.spk.nrm.eqn(info, model.info)
        # now do normalization based on specification in def file
        if (info$version>=2)
          {
            if (any(info$norm.type=='spike')) # spike based
              {
                stream.log('Do spike based normalization.\n', fname=log.fname)
                RG <- spike.normalize(RG, info, model.info)
              }
            else if (!any(info$norm.type == 'NA')) # classical
              {
                stream.log('Do classical normalization.\n', fname=log.fname)
                RG <- do.normalize(RG, info)
              }
            else
              stream.log('No normalization (data taken as is).\n', fname=log.fname)
          }
        else # only classical in old version.
          {
            if (info$norm.type != 'NA') # classical
              {
                stream.log('Do classical normalization.\n', fname=log.fname)
                RG <- do.normalize(RG, info)
              }
            else
              stream.log('No normalization (data taken as is).\n', fname=log.fname)
          }
        # now we write the raw data into a normalized file format
        if(info$datoutfnam!='NA')
          {
            stream.log('Write raw data to files.\n', fname=log.fname)
            file.out.RG(info, RG)
          }
        if(info$unbalanced) # later extension: provide analysis in that case as well.
          {
            stream.log('Unbalanced experiment can not be analysed. Returning RG and info.\n', fname=log.fname)
            ret <- list(info=info, RG=RG)
            return(ret)
          }
        # next step: parse info and remove the replicate on slide effect if it is 1
        # otherwise we get problems with 0 degrees of freedom!
        n.effects <- length(info$effects)
        if(info$effects[n.effects] ==1)
          {
            info$effects <- info$effects[1:(n.effects-1)]
            info$effnams <- info$effnams[1:(n.effects-1)]
            info$randeffs <- info$randeffs[1:(n.effects-1)]
            RG$experiments <- RG$experiments[1:(n.effects-1)]
            stream.log('Redoing model.info.\n', fname=log.fname)
            model.info <- info.2.yasmaeqns(info)
            if(info$version >= 2) # then add the spike model equations
              model.info <- get.spk.nrm.eqn(info, model.info)
          }
        # next step is to convert the RG structure into array data.
        if (info$dologarray)
          {
            stream.log('Convert RG data to log ratios.\n', fname=log.fname)
            aov.a <- rg2log.array(RG, log.ratio=T)
          }
        else
          {
            stream.log('Convert RG *log data* to log ratios (no log taken).\n', fname=log.fname)
            aov.a <- rg2array(RG)
            # we do not want varieties hence we remove them by subtratction
            rest.dims <- dim(aov.a)[-1]
            ao.nams <- names(dim(aov.a))
            tmp.a <- array(aov.a, c(2, prod(rest.dims)))
            aov.a <- array(tmp.a[1,]-tmp.a[2,], rest.dims)
            dim(aov.a) <- rest.dims
            names(dim(aov.a)) <- ao.nams[-1]
          }
        if (info$mom.stab)
          {
            stream.log('Removing first and second moment for numerical stability.\n', fname=log.fname)
            tmp.rt <- data.rem.g.moments(aov.a, info)
            aov.a <- tmp.rt$av.data
            info <- tmp.rt$info
          }
        stream.log('Calculating ANOVA and write table and variance components.\n', fname=log.fname)
        # now we specify the bfaov object that is needed for all further steps
        av <- do.anova(model.info, aov.a)
        # write the anova and reml variance table according to info
        aov.tab.varcomp.2.file(av, info, model.info)
        # get number of all comparisons to adjust p-values
        stream.log('Getting number of comparisons\n', fname=log.fname)
        n.comps <- get.n.comps(RG, info)
        if(info$do.variety)
          {
            stream.log('Ranking by Variety.\n', fname=log.fname)
            var.tab <- do.rank.variety(av, RG, model.info, info, n.comps=n.comps)
          }
        else
          var.tab <- list()
        if(info$do.anova.ranking)
          {
            # generate ANOVA based rank list
            stream.log('ANOVA ranking. \n', fname=log.fname)
            aov.tab <- fspma.aovrnk(info, model.info, aov.a, n.comps=n.comps)
          }
        else
          aov.tab <- list()
        if(!any(is.na(info$contrast)))
          {
            # write gene rank lists using the av object, RG, model.info and info
            stream.log('Contrast based ranking \n', fname=log.fname)
            ctr.tabs <- contrast.rank.2.file(av, RG, model.info, info, n.comps=n.comps, log.fname)
          }
        else
          ctr.tabs <- list()
        ret <- list(info=info, RG=RG, model.info=model.info, aov.a=aov.a, av=av,
                    plt.tabs=c(ctr.tabs, aov.tab, var.tab))
      }
    stream.log('Processing completed.\n', fname=log.fname)
    ret
  }

fspmaRG.2.RG <- function(fspmaobj)
  # extract RG compatible data from fspmaobj (fspmaobj can either be a
  # fspma object OR a fspma RG object (i.e. a fspmaobj$RG object)
  {
    if (any('RG'==names(fspmaobj))) # then this is a fspma object
      f.RG <- fspmaobj$RG
    else
      f.RG <- fspmaobj
    RG <- list(genes=f.RG$genes, R=f.RG$R, Rb=f.RG$Rb, G=f.RG$G,
               Gb=f.RG$Gb, x=f.RG$x, y=f.RG$y, experiments=f.RG$experiments)
    RG
  }

adj.RG <- function(RG, fspmaRG)
  # adjust RG (size of fspmaRG is adjusted to RG)
  # an error is raised if the number of columns in
  # RG differ the number of columns in fspmaRG
  # since this can not be corrected. We can only correct for
  # excess genes in fspmaRG that were removed.
  {
    # first we search for corresponding genes.
    gns <- RG$genes
    fgns <- fspmaRG$genes
    if((length(gns) != length(unique(gns))) | (length(fgns) != length(unique(fgns))))
      stop('Non unique gene ids - please convert RG manually!')
    allidgns <- c(1: length(gns))
    nl <- length(fgns)
    n <- 1
    ginfRG <- c()
    ginRG <- c()
    while(n <=nl)
      {
        isin <- fgns[n] == gns
        if (any(isin))
          {
            ginfRG <- c(ginfRG, n)
            ginRG <- c(ginRG, allidgns[isin])
          }
        n <- n+1
      }
    # now we use that information to reorder all gene dependent
    # measures in both RG structures 
    all.fspmaRGnams <- c('R', 'Rb', 'G', 'Gb', 'Flg', 'x', 'y', 'slideno', 'gridno')
    nl<- length(all.fspmaRGnams)
    n <- 1
    while (n <=nl)
      {
        cmp <- all.fspmaRGnams[n]
        if (any(cmp==names(RG)))
          RG[[cmp]] <- RG[[cmp]][ginRG,]
        if (any(cmp==names(fspmaRG)))
          fspmaRG[[cmp]] <- fspmaRG[[cmp]][ginfRG,]
        n <- n+1
      }
    # finally we also rearrange the genes
    RG$genes <- RG$genes[ginRG]
    fspmaRG$genes <- fspmaRG$genes[ginfRG]
    # and return both modified objects
    list(RG=RG, fspmaRG=fspmaRG)
  }

RG.2.fspmaRG <- function(RG, fspmaobj)
  # extract RG compatible data from fspmaobj (fspmaobj can either be a
  # fspma object OR a fspma RG object (i.e. a fspmaobj$RG object)
  # remark: if fspmaobj is a genuine fspma object (and not only a fspma RG)
  # fspmaobj settings override the move.2.log parameter!
  {
    have.fspma.obj <- (any('RG'==names(fspmaobj))) # then this is a fspma object
    if (have.fspma.obj)
      {
        f.RG <- fspmaobj$RG
      }
    else
      f.RG <- fspmaobj
    # first step: adjust both RG's to a common gene list
    newRGs <- adj.RG(RG, f.RG)
    RG <- newRGs$RG
    f.RG <- newRGs$fspmaRG
    # now we merge data (from RG all existing entries go to fRG)
    my.nams <- names(f.RG)
    have.nams <- names(RG)
    n.nams=length(have.nams)
    n <- 1
    while(n<=n.nams)
      {
        if (any(have.nams[n]==my.nams)) # found in foreign RG
          {
            f.RG[have.nams[n]] <- RG[have.nams[n]]
          }
        else
          {
            cat(sprintf('Warning - incompatible entry in RG: RG$%s unknown in fspma and ignored.',
                        have.nams[n]));
          }
        n <- n+1
      }
    # now f.RG is a fspma RG object and we need to find out what to do with it
    if (have.fspma.obj) # have genuine fspma object
      {
        fspmaobj$RG <- f.RG
        if (!check.geneids(fspmaobj$info, fspmaobj$RG)) # possibly genes removed from original RG!
          {
            cat('Gene names mismatch in fspmaobj resolved.\n')
            fspmaobj$info$geneids <- fspmaobj$RG$genes
          }
      }
    else                # have fspma RG object
      fspmaobj <- f.RG
    fspmaobj
  }

fspma.RG2MA <- function(fspma.obj)
  {
    RG <- fspma.obj$RG
    to.log <- fspma.obj$info$dologarray
    if (to.log)
      a <- rg2log.array(RG)
    else
      a <- rg2array(RG)
    ndim <- length(dim(a))
    rest.dims <- dim(a)[-1]
    ass.cmd <- paste('', sep='', collapse='',
                     c('R <- a[1', paste('', rep(', ', ndim-1), sep='', collapse=''), ']'))
    eval(parse(text=ass.cmd))
    ass.cmd <- paste('', sep='', collapse='',
                     c('G <- a[2', paste('', rep(', ', ndim-1), sep='', collapse=''), ']'))
    eval(parse(text=ass.cmd))
    dim(R) <- rest.dims
    dim(G) <- rest.dims
    A <- R + G
    M <- R - G
    dim(A) <- dim(R)
    dim(M) <- dim(R)
    names(dim(A)) <- names(dim(a))[-1]
    names(dim(M)) <- names(dim(a))[-1]
    list(A=A, M=M)
  }

fspma.MA2RG <- function(M, A, log2exp = F)
  {
    R <- (M + A)/2
    G <- (A - M)/2
    
    if (log2exp)
      {
        R <- 2^R
        G <- 2^G
      }
    dim(R) <- dim(M)
    dim(G) <- dim(M)
    list(R=R, G=G)
  }

fspma.avchannel <- function(fspma.obj, contrast=NULL, statenames=NULL, log2exp=F)
  {
    # this function obtaines the average channel values from a fspma.obj
    # it allows to specify a contrast. In that case the function maps the
    # average log ratios in each effect via contrasts back to virtual
    # channel values that correspond to the average log ratios obtained
    # from the contrast. In case one uses contrast one can define the state names
    # that should be used to describe the average channel values that result from it.
    # Default is state.1 and state.2. log2exp allows to obtain expression values from
    # the log values used as default.
    if (!is.null(contrast) & is.null(statenames))
      statenames <- c('diff')
    #print(statenames)
    obj.ents <- names(fspma.obj)
    #print(obj.ents)
    if (!any(obj.ents == 'model.info'))
      stop('Before plotting, one must run the analysis with fspma.wrapper')
    MA <- fspma.RG2MA(fspma.obj)
    rank.eq <- fspma.obj$model.info$eq.eff
    M.means <- ou.int.diff(MA$M, rank.eq)
    A.means <- ou.int.diff(MA$A, rank.eq)
    if (!is.null(contrast))
      {
        # now we calculate average using the contrast
        nmns <- nrow(M.means)
        multcols <- rep(1, nmns)
        dim(contrast) <- c(1, length(contrast))
        # first reweight contrast such that we get an average.
        idpos <- contrast>0
        if (any(idpos))
          {
            contrast[idpos] <- contrast[idpos]/sum(contrast[idpos])
          }
        else
          stop('Wrong contrast definition')
        idneg <- contrast<0
        if (any(idneg))
          {
            contrast[idneg] <- contrast[idneg]/abs(sum(contrast[idneg]))
          }
        else
          stop('Wrong contrast definition')
        M.means <- apply(M.means*contrast[multcols,], 1, sum)
        A.means <- apply(A.means*contrast[multcols,], 1, sum)
      }
    # now we move M.means and A.means to RG
    av.RG <- fspma.MA2RG(M.means, A.means, log2exp)
    # finally we add genes and dimension names to the object
    if (!is.null(statenames))
      {
        attr(av.RG$R, 'colnames') <- paste(statenames, '.av.R', sep='')
        attr(av.RG$G, 'colnames') <- paste(statenames, '.av.G', sep='')
      }
    else # we have a non contrast average and use the rank names as colnames)
      {
        attr(av.RG$R, 'colnames') <- paste(fspma.obj$info$rank.names, '.av.R', sep='')
        attr(av.RG$G, 'colnames') <- paste(fspma.obj$info$rank.names, '.av.G', sep='')
      }
    all.names <- c('gene.names', attr(av.RG$R, 'colnames'), attr(av.RG$G, 'colnames'));
    av.fspma.RG <- data.frame(I(fspma.obj$info$geneids), av.RG$R, av.RG$G)
    names(av.fspma.RG) <- all.names
    #tmpRG <- log.array2rg(all.means)
    # and to array representation.
    #marray <- rg2array(tmpRG)
    av.fspma.RG
  }

fspma.av.RG2file <- function(fspma.av.RG, filename='average.channels.tsv')
  {
    write.table(fspma.av.RG, file=filename, row.names=F, sep='\t', quote=F)
  }

fspma.av.scattp <- function(fspma.av.RG, statename, filename='NA', pl.title='Scatter Plot Average Expression',
                            x.leg='Average log channel value in G', y.leg='Average log channel value in R',
                            nsamples=50, lg.diff=2, ndiff.smpls=5, plt.lwd=1, n.sz=0.5, ud.sz=2, col.frst=T,
                            p.dim=7.0, leg.pos='bottomright', leg.bty='n')
  {
    # special case: 2 statenames will plot the log ratio of the second (y axis) over the log
    # ratio of the first (x axis). remember at calling to adjust x.leg and y.leg.
    # this addition helps to visualise affymetrix (other scatter plots look rather uninteresting!!)
    # our construction of the log.diff boundaries requires to divide lg.diff by sqrt(2).
    # (since we use lg.diff as normal
    # distance to the 45 degree line)
    if (!is.null(lg.diff))
      lg.diff <- lg.diff/sqrt(2)
    # first we get the corresponding columns for the plot:
    colnms <- names(fspma.av.RG)
    gene.nams <- fspma.av.RG[,1]
    id.R <- c()
    id.G <- c()
    n.states <- length(statename)
    if (n.states < 1 | n.states > 2)
      stop('One or two statenames required!')
    n <- 1
    while(n<=n.states)
      {
        id.R <- c(id.R, grep(paste(statename[n], ".av.R", sep="", collapse=""), colnms, fixed=T))
        if (any(id.R))
          R.vals <- fspma.av.RG[,id.R]
        else
          stop(sprintf('Name %s not found in column names', paste(statename, ".av.R", sep="", collapse="")))
        id.G <- c(id.G, grep(paste(statename[n], ".av.G", sep="", collapse=""), colnms, fixed=T))
        if (any(id.G))
          G.vals <- fspma.av.RG[,id.G]
        else
          stop(sprintf('Name %s not found in column names', paste(statename, ".av.G", sep="", collapse="")))
        n <- n+1
      }
    if (n.states==2)
      {
        vals.diff <- R.vals-G.vals # move to log ratios
        G.vals <- vals.diff[,1]
        R.vals <- vals.diff[,2]
      }
    # calculate average differential expression in that effect
    d.ex <- R.vals-G.vals
    n.vals <- length(d.ex)
    srt.res <- sort(d.ex, index.return=T, decreasing = T)
    # separate data: up. down and none. the latter is subsampled to nsample entries
    up.dx <- srt.res$ix[1:ndiff.smpls]
    dn.dx <- srt.res$ix[(n.vals-ndiff.smpls+1):n.vals]
    non.dx <- srt.res$ix[ndiff.smpls:(n.vals-ndiff.smpls)]
    non.dx <- sample(non.dx, nsamples)
    G.p <- G.vals[c(up.dx, dn.dx, non.dx)]
    R.p <- R.vals[c(up.dx, dn.dx, non.dx)]
    # get coordinates for the two +/- lg.diff lines and maximum plot coordinates
    xmin <- min(G.p)
    xmax <- max(G.p)
    if (!is.null(lg.diff)) # the we construct valid line coordinates
      dxy <- lg.diff/sqrt(2)
    else
      dxy <- 0
    xpmin <- xmin-dxy-0.5
    xpmax <- xmax+dxy+0.5
    xminl <- xmin+dxy
    yminl <- xmin-dxy
    xminu <- xmin-dxy
    yminu <- xmin+dxy
    xmaxl <- xmax+dxy
    ymaxl <- xmax-dxy
    xmaxu <- xmax-dxy
    ymaxu <- xmax+dxy

    xll <- c(xminl, xmaxl)
    yll <- c(yminl, ymaxl)
    xul <- c(xminu, xmaxu)
    yul <- c(yminu, ymaxu)

    ypmin <- min(c(R.vals, yminl))-0.5
    ypmax <- max(c(R.vals, ymaxu))+0.5
    # open plot device
    if (filename=='new') # to new plot screen
      x11()
    else if (filename != 'NA')                # to eps file
      postscript(filename, width = p.dim, height = p.dim,
                 horizontal = FALSE, onefile = FALSE, paper = "special")
    plot(G.vals[non.dx], R.vals[non.dx], type='p', ylim=c(ypmin, ypmax), xlim=c(xpmin, xpmax),
         col=1, pch=1, cex=n.sz, main=pl.title, xlab=x.leg, ylab=y.leg)
    if (!is.null(lg.diff)) # the add the lines that mark the lg.diff boundary
      {
        lines(xll, yll, col=1, lty=1, lwd=plt.lwd)
        lines(xul, yul, col=1, lty=1, lwd=plt.lwd)
      }
    # points(G.vals[non.dx], R.vals[non.dx], pch=1, col=1, cex=n.sz)
    # now construct legend and symbols for up and down.
    dud <- c(up.dx, dn.dx)
    n.dud <- length(dud)
    leg.txt <- c()
    cols <- c()
    pc <- c()
    for (n in 1:n.dud)
      {
        # first get symbol style
        ret <- get.colstyle(n, col.frst)
        x <- G.vals[dud[n]]
        y <- R.vals[dud[n]]
        points(x, y, col=ret$col.no, pch=ret$stl.no, cex=ud.sz)
        leg.txt <- c(leg.txt,  gene.nams[dud[n]])
        cols <- c(cols, ret$col.no)
        pc <- c(pc, ret$stl.no)
      }
    if (!is.na(as.numeric(leg.pos)))
      {
        leg.x=leg.pos[1]
        leg.y=leg.pos[2]
        legend(leg.x, leg.y, leg.txt, col=cols, pch=pc, bty=leg.bty)
      }
    else
      {
        legend(leg.pos, leg.txt, col=cols, pch=pc, bty=leg.bty)
      }
    if (filename!='NA' & filename!='new') # close plot device. (eps file)
      dev.off()
  }

fspma.maplot <- function(fspma.obj, slideno=1, pl.file='NA', pl.title='M-A Plot',
                         have.spike='no', nsmpls=50, xlab='Log Amplitude',
                         ylab='Log Ratio', cex=0.5, col=2, leg.pos='bottomright', leg.bty='n')
  # note that have.spike options are 'no' or 'yes'
  # in case of yes RG should contain the unnormalized data
  # in case of norm RG should be the normalized data. (otherwise there is a mismatch!)
  {
    RG <- fspma.obj$RG
    info <- fspma.obj$info
    have.spike <- (have.spike=='yes') & !is.null(info$spikegenes$names) # we set spikes='yes' and have them
    norm.data <- any('model.info'==names(fspma.obj))
    if (norm.data)
      model.info <- fspma.obj$model.info
    else
      {
        model.info <- info.2.yasmaeqns(info)
        if(info$version >= 2) # then add the spike model equations
          model.info <- get.spk.nrm.eqn(info, model.info)
      }
    if(any(names(RG)=='slideno'))
      {
        # open window or file
        if(pl.file=='new')
          x11()
        else if (pl.file != 'NA')
          postscript(pl.file, width = 8.0, height = 6.0,
                     horizontal = FALSE, onefile = FALSE, paper = "special")
        # extract genes
        idx <- RG$slideno==slideno
        R.i <- RG$R[idx]-RG$Rb[idx]
        G.i <- RG$G[idx]-RG$Gb[idx]
        if (info$dologarray)
          {
            R.i[R.i < fspma.eps] <- fspma.eps
            G.i[G.i < fspma.eps] <- fspma.eps
            R.i <- log2(R.i)
            G.i <- log2(G.i)
          }
        A <- as.vector(R.i+G.i)
        M <- as.vector(R.i-G.i)
        n.s <- min(nsmpls, length(A))
        A <- A[1:n.s]
        xmin <- min(A)
        xmax <- max(A)
        ymin <- min(M)
        ymax <- max(M)
        M <- M[1:n.s]
        # extract spikes
        spk.plt <- F
        if (have.spike)
          {
            if (!norm.data) # we have unnormalized data and spikes
              {
                spikes <- RGspike.2.spikeframe(RG, info, model.info, slideno=slideno)
                A.s <- as.vector(spikes$A)
                n.s <- min(nsmpls, length(A.s))
                A.s <- A.s[1:n.s]
                M.s <-  as.vector(spikes$M[1:n.s])+log(as.vector(spikes$t.rt[1:n.s]))
                xmin <- min(c(xmin, A.s))
                xmax <- max(c(xmax, A.s))
                ymin <- min(c(ymin, M.s))
                ymax <- max(c(ymax, M.s))
                spk.plt <- T
              }
            else # we have normalized data and spikes
              {
                spikes <- RGnormalized.spike.2.spikeframe(RG, slideno=slideno)
                A.s <- as.vector(spikes$A)
                n.s <- min(nsmpls, length(A.s))
                A.s <- A.s[1:n.s]
                M.s <-  as.vector(spikes$M[1:n.s])+log(as.vector(spikes$t.rt[1:n.s]))
                xmin <- min(c(xmin, A.s))
                xmax <- max(c(xmax, A.s))
                ymin <- min(c(ymin, M.s))
                ymax <- max(c(ymax, M.s))
                spk.plt <- T
              }
          }
        plot(A, M, xlim=c(xmin, xmax), ylim=c(ymin, ymax), type='p', main=pl.title,
             col=col, pch=1, cex=cex, xlab=xlab, ylab=ylab)
        if (spk.plt)
          {
            points(A.s, M.s, pch=3, col=4)
            legtext <- c('genes', 'spikes')
            cols <- c(2, 4)
            pc <- c(1, 3)
            if (!is.na(as.numeric(leg.pos)))
              {
                leg.x=leg.pos[1]
                leg.y=leg.pos[2]
                legend(leg.x, leg.y, legtext, col=cols, pch=pc, bty=leg.bty)
              }
            else
              {
                legend(leg.pos, legtext, col=cols, pch=pc, bty=leg.bty)
              }
          }
        if (pl.file!='NA' & pl.file != 'new') # close plot device.
            dev.off()
      }
  }

get.colstyle <- function(n.plt, col.frst=T)
  {
    all.cols <- 8
    all.styl <- 6
    if (col.frst)
      {
        col.no <- ((n.plt-1) %% all.cols)+1
        stl.no <- ((n.plt-1)%/% all.cols)+1
      }
    else
      {
        col.no <- ((n.plt-1) %% all.styl)+1
        stl.no <- ((n.plt-1) %/% all.styl)+1
      }
     list(col.no=col.no, stl.no=stl.no)
  }
        

fspma.rankplot <- function(fspma.obj, list.nam, no.genes=5, pl.title='Rank List',
                           pl.file='NA', plt.lwd=2, col.frst=T, leg.pos='bottomright',
                           leg.bty='n', def.xlabel='default', def.ylabel='default')
  {
    # plots the normalized log expression values of no.genes top raked genes into a window
    # or into a file.
    # note that legheight can be 'up', 'mid' or 'down'
    # The table to be plotted is selected via its name.
    # column names of data frame: c('dx', 'g.nams', 'p.vals')
    # or c('dx', 'g.nams', 'p.val', 'is.up', mn.nams)
    rank.lists <- fspma.obj$plt.tabs
    rank.nams <- names(rank.lists)
    id.this <- grep(list.nam, rank.nams, fixed=T)
    if(any(id.this))
      {
        data.tab <- rank.lists[[id.this]]
        tab.names <- names(data.tab)
        n.genes <- min(nrow(data.tab), no.genes)
        leg.name <- data.tab[1:n.genes, 'g.nams']
        n.cols <- ncol(data.tab)
        #print(tab.names)
        if (tab.names[4]=='is.up') # then we have a contrats
          data.cols <- c(5:n.cols)
        else # we have a anova rank table
          data.cols <- c(4:n.cols)
        ymin <- min(as.vector(data.tab[1:n.genes, data.cols]))
        ymax <- max(as.vector(data.tab[1:n.genes, data.cols]))
        ydif <- ymax-ymin
        ymin <- ymin-0.05*ydif
        ymax <- ymax+0.05*ydif
        if (2>=(n.pt <- length(data.cols)))
          {
            # plot points
            x <- c(1:n.pt)
            xmin <- 0
            xmax <- n.pt*n.genes
            y <- as.vector(data.tab[1, data.cols])
            if (pl.file=='new')
              x11()
            else if (pl.file != 'NA')
              postscript(pl.file, width = 8.0, height = 6.0,
                         horizontal = FALSE, onefile = FALSE, paper = "special")
            # set x and y label
            if (def.xlabel == 'default')
              myxlab= 'sample index'
            else
              myxlab=def.xlabel
            if (def.ylabel == 'default')
              myylab= 'log ratio'
            else
              myylab=def.ylabel
            
            plot(x, y, ylim=c(ymin, ymax), xlim=c(xmin, xmax), col=1, pch=1,
                 main=pl.title, xlab=myxlab, ylab=myylab)
            all.col <- 1
            all.pc <- 1
            for (n in c(2:n.genes))
              {
                y <- as.vector(data.tab[n, data.cols])
                ret <- get.colstyle(n, col.frst)
                x <- n.pt*(n-1)+c(1:n.pt)
                points(x, y, col=ret$col.no, pch=ret$stl.no)
                all.col <- c(all.col, ret$col.no)
                all.pc <- c(all.pc, ret$stl.no)
              }
            if (!is.na(as.numeric(leg.pos)))
              {
                leg.x=leg.pos[1]
                leg.y=leg.pos[2]
                legend(leg.x, leg.y, leg.name, pch=all.pc, col=all.col, bty=leg.bty)
              }
            else
              {
                legend(leg.pos, leg.name, pch=all.pc, col=all.col, bty=leg.bty)
              }
            if (pl.file!='NA' & pl.file != 'new')
              dev.off()
          }
        else
          {
            # plot time course
            x <- as.vector(c(1:n.pt))
            xmin <- 0
            xmax <- n.pt
            y <- as.vector(data.tab[1, data.cols])
            if (pl.file=='new')
              x11()
            else if (pl.file != 'NA')
              postscript(pl.file, width = 8.0, height = 6.0,
                         horizontal = FALSE, onefile = FALSE, paper = "special")
            # set x and y label
            if (def.xlabel == 'default')
              myxlab= 'time point'
            else
              myxlab=def.xlabel
            if (def.ylabel == 'default')
              myylab= 'log ratio'
            else
              myylab=def.ylabel

            plot(x, y, type='l', ylim=c(ymin, ymax), xlim=c(xmin, xmax), col=1, lty=1, lwd=plt.lwd,
                 main=pl.title, xlab=myxlab, ylab=myylab)
            all.col <- 1
            all.lt <- 1
            for (n in c(2:n.genes))
              {
                y <- as.vector(data.tab[n, data.cols])
                ret <- get.colstyle(n, col.frst)
                lines(x, y, col=ret$col.no, lty=ret$stl.no, lwd=plt.lwd)
                all.col <- c(all.col, ret$col.no)
                all.lt <- c(all.lt, ret$stl.no)
              }
            if (!is.na(as.numeric(leg.pos)))
              {
                leg.x=leg.pos[1]
                leg.y=leg.pos[2]
                legend(leg.x, leg.y, leg.name, lty=all.lt, col=all.col, bty=leg.bty)
              }
            else
              {
                legend(leg.pos, leg.name, lty=all.lt, col=all.col, bty=leg.bty)
              }
            if (pl.file!='NA' & pl.file != 'new')
              dev.off()
          }
      }
    else
      cat(sprintf('List %s not found. Available rank lists:%s\n',list.nam,
                  paste("", names(rank.lists), sep=" ", collapse=""))) 
  }

# fspma2vsnarray converts a fspma object into a vsn compatible array, where every
# column represents all expression values per gene. Note that vsnarray returns
# all channels in one matrix. For one channel arrays, only the R channel is returned.
fspma2vsnarray <- function(fspma.obj)
  {
    # stack RG
    RG <- stack.reps(fspma.obj$RG, fspma.obj$info)
    # get values - note that Rb and Gb are 0 if not available and can
    # always be subtracted safely
    if (info.single(fspma.obj$info))
      {
        vsnarray <- RG$R-RG$Rb
      }
    else
      {
        vsnarray <- cbind(RG$R-RG$Rb, RG$G-RG$Gb)
      }
    vsnarray
  }

# vsnarray2fspmaobj exchanges expression patterns in a fspma RG object with data taken from
# a vsn compatible array.
vsnarray2fspma <- function(fspma.obj, vsnarray)
  {
    # get stacked RG
    #print('1')
    #print(dim(vsnarray))
    RG <- stack.reps(fspma.obj$RG, fspma.obj$info)
    #print('2')
    #print(names(RG))
    # alter normalised values
    if (info.single(fspma.obj$info))
      {
        #print('3.1')
        RG$R <- vsnarray
      }
    else
      {
        #print('3.2.1')
        ncolsRG <- ncol(RG$R)
        #print(ncolsRG)
        #print('3.2.2')
        RG$R <- vsnarray[,1:ncolsRG]
        #print('3.2.3')
        RG$G <- vsnarray[,ncolsRG+1:ncolsRG]
        #print('3.2.4')
      }
    #print('4')
    # unstack RG
    RG <- unstack.reps(RG, fspma.obj$info)
    #print('5')
    # and write back to fspma.obj
    fspma.obj <- RG.2.fspmaRG(RG, fspma.obj)
    #print('6')
    fspma.obj
  }


